# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import subprocess
import shutil
import gzip
from datetime import datetime

from spectre_core.config import (
    get_spectre_data_dir_path, get_batches_dir_path, TimeFormat, trim_spectre_data_dir_path
)

from enum import Enum

class CallistoInstrumentCode(Enum):
    """e-Callisto network station codes."""
    ALASKA_ANCHORAGE     = "ALASKA-ANCHORAGE"
    ALASKA_COHOE         = "ALASKA-COHOE"
    ALASKA_HAARP         = "ALASKA-HAARP"
    ALGERIA_CRAAG        = "ALGERIA-CRAAG"
    ALMATY               = "ALMATY"
    AUSTRIA_KRUMBACH     = "AUSTRIA-Krumbach"
    AUSTRIA_MICHELBACH   = "AUSTRIA-MICHELBACH"
    AUSTRIA_OE3FLB       = "AUSTRIA-OE3FLB"
    AUSTRIA_UNIGRAZ      = "AUSTRIA-UNIGRAZ"
    AUSTRALIA_ASSA       = "Australia-ASSA"
    BIR                  = "BIR"
    CROATIA_VISNJAN      = "Croatia-Visnjan"
    DENMARK              = "DENMARK"
    EGYPT_ALEXANDRIA     = "EGYPT-Alexandria"
    EGYPT_SPACEAGENCY    = "EGYPT-SpaceAgency"
    FINLAND_SIUNTIO      = "FINLAND-Siuntio"
    FINLAND_KEMPELE      = "Finland-Kempele"
    GERMANY_DLR          = "GERMANY-DLR"
    GLASGOW              = "GLASGOW"
    GREENLAND            = "GREENLAND"
    HUMAIN               = "HUMAIN"
    HURBANOVO            = "HURBANOVO"
    INDIA_GAURI          = "INDIA-GAURI"
    INDIA_OOTY           = "INDIA-OOTY"
    INDIA_UDAIPUR        = "INDIA-UDAIPUR"
    JAPAN_IBARAKI        = "JAPAN-IBARAKI"
    KASI                 = "KASI"
    MEXART               = "MEXART"
    MEXICO_FCFM_UANL     = "MEXICO-FCFM-UANL"
    MEXICO_LANCE_A       = "MEXICO-LANCE-A"
    MEXICO_LANCE_B       = "MEXICO-LANCE-B"
    MONGOLIA_UB          = "MONGOLIA-UB"
    MRO                  = "MRO"
    MRT3                 = "MRT3"
    MALAYSIA_BANTING     = "Malaysia-Banting"
    NORWAY_EGERSUND      = "NORWAY-EGERSUND"
    NORWAY_NY_AALESUND   = "NORWAY-NY-AALESUND"
    NORWAY_RANDABERG     = "NORWAY-RANDABERG"
    POLAND_GROTNIKI      = "POLAND-Grotniki"
    ROMANIA              = "ROMANIA"
    ROSWELL_NM           = "ROSWELL-NM"
    SPAIN_PERALEJOS      = "SPAIN-PERALEJOS"
    SSRT                 = "SSRT"
    SWISS_HB9SCT         = "SWISS-HB9SCT"
    SWISS_HEITERSWIL     = "SWISS-HEITERSWIL"
    SWISS_IRSOL          = "SWISS-IRSOL"
    SWISS_LANDSCHLACHT   = "SWISS-Landschlacht"
    SWISS_MUHEN          = "SWISS-MUHEN"
    TRIEST               = "TRIEST"
    TURKEY               = "TURKEY"
    UNAM                 = "UNAM"
    URUGUAY              = "URUGUAY"
    USA_BOSTON           = "USA-BOSTON"


def _get_batch_name(
    station: str, 
    date: str, 
    time: str, 
    code: str
) -> str:
    """
    Create a standardised batch file name for a `spectre` batch file.

    :param station: Station name.
    :param date: Observation date in 'YYYYMMDD' format.
    :param time: Observation time in 'HHMMSS' format.
    :param code: Numeric instrument code.
    :return: Formatted batch file name.
    """
    dt = datetime.strptime(f"{date}T{time}", "%Y%m%dT%H%M%S")
    formatted_time = dt.strftime(TimeFormat.DATETIME)
    return f"{formatted_time}_callisto-{station.lower()}-{code}.fits"


def _get_batch_components(
    gz_path: str
) -> list[str]:
    """
    Extract station, date, time, and instrument code from the `.fit.gz` file name.

    :param gz_path: Path to the `.fit.gz` file.
    :return: List of [station, date, time, instrument_code].
    :raises ValueError: If file format is invalid.
    """
    file_name = os.path.basename(gz_path)
    if not file_name.endswith(".fit.gz"):
        raise ValueError(f"Invalid file extension: {file_name}. Expected .fit.gz")
    file_base_name = file_name.rstrip(".fit.gz")
    parts = file_base_name.split('_')
    if len(parts) != 4:
        raise ValueError("Invalid file name format. Expected '[station]_[date]_[time]_[code].fit.gz'")
    return parts


def _get_batch_path(
    gz_path: str
) -> str:
    """
    Generate the full path for the `spectre` batch file.

    :param gz_path: Path to the raw `.fit.gz` file.
    :return: Path to the corresponding batch file.
    """
    station, date, time, code = _get_batch_components(gz_path)
    batch_name = _get_batch_name(station, date, time, code)
    batch_start_time = batch_name.split("_")[0]
    dt = datetime.strptime(batch_start_time, TimeFormat.DATETIME)
    batch_dir = get_batches_dir_path(year=dt.year, month=dt.month, day=dt.day)
    os.makedirs(batch_dir, exist_ok=True)
    return os.path.join(batch_dir, batch_name)


def _unzip_file_to_batches(
    gz_path: str
) -> str:
    """
    Decompress a `.fit.gz` file and save it as a `.fits` batch file.

    :param gz_path: Path to the `.fit.gz` file.
    :return: The file path of the newly created batch file, relative to the mounted volume.
    """
    fits_path = _get_batch_path(gz_path)
    with gzip.open(gz_path, "rb") as f_in, open(fits_path, "wb") as f_out:
        shutil.copyfileobj(f_in, f_out)
        return trim_spectre_data_dir_path(f_out.name)


def _unzip_to_batches(
    tmp_dir: str
) -> list[str]:
    """
    Decompress all `.gz` files in a temporary directory and save them as `spectre`
    batch files.

    :param tmp_dir: Path to the temporary directory containing `.gz` files.
    :return: A list of file names of all newly created batch files, relative to the mounted volume.
    """
    batch_file_names = []
    for entry in os.scandir(tmp_dir):
        if entry.is_file() and entry.name.endswith(".gz"):
            batch_file_names.append( _unzip_file_to_batches(entry.path) )
            os.remove(entry.path)
    shutil.rmtree(tmp_dir)
    return batch_file_names
            


def _wget_callisto_data(
    instrument_code: str, 
    year: int, 
    month: int, 
    day: int, 
    tmp_dir: str
) -> None:
    """
    Download raw `.fit.gz` files from the e-Callisto network using `wget`.

    :param instrument_code: Instrument code for filtering files.
    :param year: Observation year.
    :param month: Observation month.
    :param day: Observation day.
    :param tmp_dir: Path to the temporary directory for downloads.
    """
    date_str = f"{year:04d}/{month:02d}/{day:02d}"
    base_url = f"http://soleil.i4ds.ch/solarradio/data/2002-20yy_Callisto/{date_str}/"
    command = [
        "wget", "-r", "-l1", "-nd", "-np",
        "-R", ".tmp", "-A", f"{instrument_code}*.fit.gz",
        "-P", tmp_dir, base_url
    ]
    subprocess.run(command, check=True)


def download_callisto_data(
    instrument_code: CallistoInstrumentCode, 
    year: int, 
    month: int, 
    day: int
) -> list[str]:
    """
    Download and decompress e-Callisto FITS files, saving them as `spectre` batch files.

    :param instrument_code: e-Callisto station instrument code.
    :param year: Year of the observation.
    :param month: Month of the observation.
    :param day: Day of the observation.
    :return: A list of file names of all newly created batch files, relative to the mounted volume.
    """
    tmp_dir = os.path.join(get_spectre_data_dir_path(), "tmp")
    # if there are any residual files in the temporary directory, remove them.
    if os.path.exists(tmp_dir):
        shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir, exist_ok=True)
    
    _wget_callisto_data(instrument_code.value, year, month, day, tmp_dir)
    return sorted( _unzip_to_batches(tmp_dir) )


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import numpy.typing as npt
from typing import Optional, Tuple, cast
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes

from spectre_core.spectrograms import TimeType
from ._base import BasePanel, XAxisType
from ._format import PanelFormat
from ._panels import PanelName, SpectrogramPanel, TimeCutsPanel, FrequencyCutsPanel


def _is_cuts_panel(
    panel: BasePanel
) -> bool:
    """Check if a panel contains spectrogram cuts.

    :param panel: The panel to check.
    :return: True if the panel is a frequency or time cuts panel, otherwise False.
    """
    return panel.name in {PanelName.FREQUENCY_CUTS, PanelName.TIME_CUTS}


def _is_spectrogram_panel(
    panel: BasePanel
) -> bool:
    """Check if a panel is a spectrogram panel.

    :param panel: The panel to check.
    :return: True if the panel is a spectrogram panel, otherwise False.
    """
    return panel.name == PanelName.SPECTROGRAM


class PanelStack:
    """Visualise spectrogram data in a stack of panels."""
    def __init__(
        self, 
        panel_format: PanelFormat = PanelFormat(),
        time_type: TimeType = TimeType.RELATIVE,
        figsize: Tuple[int, int] = (10, 10)
    ) -> None:
        """Initialize an instance of `PanelStack`.

        :param panel_format: Formatting applied across all panels in the stack. Defaults to `PanelFormat()`.
        :param time_type: The type of time assigned to spectrograms, defaults to `TimeType.RELATIVE`.
        :param figsize: The size of the `matplotlib` figure as (width, height). Defaults to (10, 10).
        """
        self._panel_format = panel_format
        self._time_type = time_type
        self._figsize = figsize

        self._panels             : list[BasePanel] = []
        self._superimposed_panels: list[BasePanel] = []
        
        self._fig: Optional[Figure] = None
        self._axs: Optional[np.ndarray] = None


    @property
    def time_type(
        self
    ) -> TimeType:
        """The type of time we assign to the spectrograms"""
        return self._time_type


    @property
    def panels(
        self
    ) -> list[BasePanel]:
        """Get the panels in the stack, sorted by their `XAxisType`."""
        return list(sorted(self._panels, key=lambda panel: panel.xaxis_type.value))


    @property
    def fig(
        self
    ) -> Figure:
        """Get the shared `matplotlib` figure for the panel stack.

        :raises ValueError: If the axes have not been initialized.
        """
        if self._fig is None:
            raise ValueError(f"An unexpected error has occured, `fig` must be set for the panel stack.")
        return self._fig


    @property
    def axs(
        self
    ) -> np.ndarray:
        """Get the `matplotlib` axes in the stack.

        :return: An array of `matplotlib.axes.Axes`, one for each panel in the stack.
        :raises ValueError: If the axes have not been initialized.
        """
        if self._axs is None:
            raise ValueError(f"An unexpected error has occured, `axs` must be set for the panel stack.")
        return np.atleast_1d(self._axs)


    @property
    def num_panels(
        self
    ) -> int:
        """Get the number of panels in the stack."""
        return len(self._panels)


    def add_panel(
        self, 
        panel: BasePanel,
        identifier: Optional[str] = None
    ) -> None:
        """Add a panel to the stack.

        :param panel: An instance of a `BasePanel` subclass to be added to the stack.
        :param identifier: An optional string to link the panel with others for superimposing.
        """
        panel.panel_format = self._panel_format
        panel.time_type    = self._time_type
        if identifier: 
            panel.identifier = identifier
        self._panels.append(panel)


    def superimpose_panel(
        self, 
        panel: BasePanel,
        identifier: Optional[str] = None
    ) -> None:
        """Superimpose a panel onto an existing panel in the stack.

        :param panel: The panel to superimpose.
        :param identifier: An optional identifier to link panels during superimposing, defaults to None
        """
        if identifier:
            panel.identifier = identifier
        panel.panel_format = self._panel_format
        self._superimposed_panels.append(panel)


    def _init_plot_style(
        self
    ) -> None:
        """Initialize the global plot style for the stack.

        This method sets `matplotlib` styles and font sizes based on the `panel_format`.
        """
        plt.style.use(self._panel_format.style)
        
        plt.rc('font', size=self._panel_format.small_size)
        
        plt.rc('axes', titlesize=self._panel_format.medium_size, 
                       labelsize=self._panel_format.medium_size)
        
        plt.rc('xtick', labelsize=self._panel_format.small_size)
        plt.rc('ytick', labelsize=self._panel_format.small_size)
        
        plt.rc('legend', fontsize=self._panel_format.small_size)
        plt.rc('figure', titlesize=self._panel_format.large_size)


    def _create_figure_and_axes(
        self
    ) -> None:
        """Create the `matplotlib` figure and axes for the panel stack.

        This initializes a figure with a specified number of vertically stacked axes.
        """
        self._fig, self._axs = plt.subplots(self.num_panels, 
                                            1, 
                                            figsize=self._figsize, 
                                            layout="constrained")


    def _assign_axes(
        self
    ) -> None:
        """Assign each axes in the figure to some panel in the stack.
        
        Axes are shared between panels with common `xaxis_type`.
        """
        shared_axes: dict[XAxisType, Axes] = {}
        for i, panel in enumerate(self.panels):
            panel.ax = self.axs[i]
            panel.fig = self._fig
            if panel.xaxis_type in shared_axes:
                panel.ax.sharex( shared_axes[panel.xaxis_type] )
            else:
                shared_axes[panel.xaxis_type] = panel.ax


    def _overlay_cuts(
        self, 
        cuts_panel: BasePanel
    ) -> None:
        """Overlay cuts onto corresponding spectrogram panels.

        For a given cuts panel, locate any spectrogram panels in the stack with a matching tag
        and add the appropriate frequency or time cuts overlay.

        :param cuts_panel: A panel containing frequency or time cuts.
        """
        for panel in self.panels:
            if _is_spectrogram_panel(panel) and panel.tag == cuts_panel.tag:
                panel = cast(SpectrogramPanel, panel)
                if cuts_panel.name == PanelName.FREQUENCY_CUTS:
                    panel.overlay_frequency_cuts( cast(FrequencyCutsPanel, cuts_panel) )
                elif cuts_panel.name == PanelName.TIME_CUTS:
                    panel.overlay_time_cuts( cast(TimeCutsPanel, cuts_panel) )


    def _overlay_superimposed_panels(
        self
    ) -> None:
        """Superimpose panels onto matching panels in the stack.

        For each superimposed panel, find a matching panel in the stack with the same name 
        and identifier. Share the axes and figure, then draw the superimposed panel. If the 
        panel contains cuts, overlay those cuts onto the corresponding spectrogram panels.
        """
        for super_panel in self._superimposed_panels:
            for panel in self._panels:
                if panel.name == super_panel.name and (panel.identifier == super_panel.identifier):
                    super_panel.ax, super_panel.fig = panel.ax, self._fig
                    super_panel.draw()
                    if _is_cuts_panel(super_panel):
                        self._overlay_cuts(super_panel)


    def show(
        self
    ) -> None:
        """Display the panel stack figure."""
        self._init_plot_style()
        self._create_figure_and_axes()
        self._assign_axes()
        
        last_panel_per_axis = {panel.xaxis_type: panel for panel in self.panels}
        
        for panel in self.panels:
            panel.draw()
            panel.annotate_yaxis()
            
            if panel == last_panel_per_axis[panel.xaxis_type]:
                panel.annotate_xaxis()
            else:
                panel.hide_xaxis_labels()
                
            if _is_cuts_panel(panel):
                self._overlay_cuts(panel)
                
        self._overlay_superimposed_panels()
        plt.show()



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Callable, cast

from spectre_core.capture_configs import (
    CaptureTemplate, CaptureMode, Parameters, Bound, PName,
    get_base_capture_template, make_base_capture_template, get_base_ptemplate,
    validate_window
)
from .gr._test import CaptureMethod
from ._receiver_names import ReceiverName
from .._spec_names import SpecName
from .._base import BaseReceiver
from .._register import register_receiver


@dataclass(frozen=True)
class Mode:
    """An operating mode for the `Test` receiver."""
    COSINE_SIGNAL_1  = "cosine_signal_1"
    TAGGED_STAIRCASE = "tagged_staircase"


@register_receiver(ReceiverName.TEST)
class Test(BaseReceiver):
    """An entirely software-defined receiver, which generates synthetic signals."""
    def _add_specs(
        self
    ) -> None:
        self.add_spec( SpecName.SAMPLE_RATE_LOWER_BOUND, 64000  )
        self.add_spec( SpecName.SAMPLE_RATE_UPPER_BOUND, 640000 )
        self.add_spec( SpecName.FREQUENCY_LOWER_BOUND  , 16000  )
        self.add_spec( SpecName.FREQUENCY_UPPER_BOUND  , 160000 )


    def _add_capture_methods(
        self
    ) -> None:
        self.add_capture_method(Mode.COSINE_SIGNAL_1 , 
                                CaptureMethod.cosine_signal_1 )
        self.add_capture_method(Mode.TAGGED_STAIRCASE, 
                                CaptureMethod.tagged_staircase)
    

    def _add_pvalidators(
        self
    ) -> None:
        self.add_pvalidator(Mode.COSINE_SIGNAL_1 , 
                            self.__get_pvalidator_cosine_signal_1()  )
        self.add_pvalidator(Mode.TAGGED_STAIRCASE, 
                            self.__get_pvalidator_tagged_staircase() )


    def _add_capture_templates(
        self
    ) -> None:
        self.add_capture_template(Mode.COSINE_SIGNAL_1 , 
                                  self.__get_capture_template_cosine_signal_1()  )
        self.add_capture_template(Mode.TAGGED_STAIRCASE, 
                                  self.__get_capture_template_tagged_staircase() )
        

    def __get_capture_template_cosine_signal_1(
        self
    ) -> CaptureTemplate:
        capture_template = get_base_capture_template( CaptureMode.FIXED_CENTER_FREQUENCY )
        capture_template.add_ptemplate( get_base_ptemplate(PName.AMPLITUDE) )
        capture_template.add_ptemplate( get_base_ptemplate(PName.FREQUENCY) )

        capture_template.set_defaults(
            (PName.BATCH_SIZE,            3.0),
            (PName.CENTER_FREQUENCY,      16000),
            (PName.AMPLITUDE,             2.0),
            (PName.FREQUENCY,             32000),
            (PName.SAMPLE_RATE,           128000),
            (PName.WINDOW_HOP,            512),
            (PName.WINDOW_SIZE,           512),
            (PName.WINDOW_TYPE,           "boxcar")
        )

        capture_template.enforce_defaults(
            PName.TIME_RESOLUTION,
            PName.TIME_RANGE,
            PName.FREQUENCY_RESOLUTION,
            PName.WINDOW_TYPE
        )


        capture_template.add_pconstraint(
            PName.SAMPLE_RATE,
            [
                Bound(
                    lower_bound=self.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND)
                )
            ]
        )
        capture_template.add_pconstraint(
            PName.FREQUENCY,
            [
                Bound(
                    lower_bound=self.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                    upper_bound=self.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
                )
            ]
        )
        return capture_template


    def __get_capture_template_tagged_staircase(
        self
    ) -> CaptureTemplate:
        capture_template = make_base_capture_template(
                PName.TIME_RESOLUTION,
                PName.FREQUENCY_RESOLUTION,
                PName.TIME_RANGE,
                PName.SAMPLE_RATE,
                PName.BATCH_SIZE,
                PName.WINDOW_TYPE,
                PName.WINDOW_HOP,
                PName.WINDOW_SIZE,
                PName.EVENT_HANDLER_KEY,
                PName.BATCH_KEY,
                PName.WATCH_EXTENSION,
                PName.MIN_SAMPLES_PER_STEP,
                PName.MAX_SAMPLES_PER_STEP,
                PName.FREQUENCY_STEP,
                PName.STEP_INCREMENT,
                PName.OBS_ALT,
                PName.OBS_LAT,
                PName.OBS_LON,
                PName.OBJECT,
                PName.ORIGIN,
                PName.TELESCOPE,
                PName.INSTRUMENT
        )

        capture_template.set_defaults(
            (PName.BATCH_SIZE,            3.0),
            (PName.FREQUENCY_STEP,        128000),
            (PName.MAX_SAMPLES_PER_STEP,  5000),
            (PName.MIN_SAMPLES_PER_STEP,  4000),
            (PName.SAMPLE_RATE,           128000),
            (PName.STEP_INCREMENT,        200),
            (PName.WINDOW_HOP,            512),
            (PName.WINDOW_SIZE,           512),
            (PName.WINDOW_TYPE,           "boxcar"),
            (PName.EVENT_HANDLER_KEY,     "swept_center_frequency"),
            (PName.BATCH_KEY,             "iq_stream"),
            (PName.WATCH_EXTENSION,       "bin")
        )

        capture_template.enforce_defaults(
            PName.TIME_RESOLUTION,
            PName.TIME_RANGE,
            PName.FREQUENCY_RESOLUTION,
            PName.WINDOW_TYPE,
            PName.EVENT_HANDLER_KEY,
            PName.BATCH_KEY,
            PName.WATCH_EXTENSION
        )

        return capture_template
    
    
    def __get_pvalidator_cosine_signal_1(
        self
    ) -> Callable[[Parameters], None]:
        def pvalidator(parameters: Parameters) -> None:
            validate_window(parameters)

            sample_rate          = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
            window_size          = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
            frequency            = cast(float, parameters.get_parameter_value(PName.FREQUENCY))

            # check that the sample rate is an integer multiple of the underlying signal frequency
            if sample_rate % frequency != 0:
                raise ValueError("The sampling rate must be some integer multiple of frequency")

            a = sample_rate/frequency
            if a < 2:
                raise ValueError((f"The ratio of sampling rate over frequency must be greater than two. "
                                f"Got {a}"))
            
            # analytical requirement
            # if p is the number of sampled cycles, we can find that p = window_size / a
            # the number of sampled cycles must be a positive natural number.
            p = window_size / a
            if window_size % a != 0:
                raise ValueError((f"The number of sampled cycles must be a positive natural number. "
                                  f"Computed that p={p}"))
        return pvalidator


    def __get_pvalidator_tagged_staircase(
        self
    ) -> Callable[[Parameters], None]:
        def pvalidator(parameters: Parameters) -> None:
            validate_window(parameters)

            freq_step            = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
            sample_rate          = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
            min_samples_per_step = cast(int, parameters.get_parameter_value(PName.MIN_SAMPLES_PER_STEP))
            max_samples_per_step = cast(int, parameters.get_parameter_value(PName.MAX_SAMPLES_PER_STEP))

            if freq_step != sample_rate:
                raise ValueError(f"The frequency step must be equal to the sampling rate")
            
            
            if min_samples_per_step > max_samples_per_step:
                raise ValueError((f"Minimum samples per step cannot be greater than the maximum samples per step. "
                                f"Got {min_samples_per_step}, which is greater than {max_samples_per_step}"))
            
        return pvalidator
        


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, overload

from spectre_core.capture_configs import (
    CaptureTemplate, CaptureMode, Parameters, Bound, PName,
    get_base_capture_template, get_base_ptemplate, OneOf,
    validate_sample_rate_with_master_clock_rate,
    validate_fixed_center_frequency, validate_swept_center_frequency
)
from .._base import BaseReceiver
from .._spec_names import SpecName


def get_pvalidator_fixed_center_frequency(
    usrp_receiver: BaseReceiver 
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_fixed_center_frequency(parameters)
        validate_sample_rate_with_master_clock_rate(parameters)
    return pvalidator


def get_pvalidator_swept_center_frequency(
    usrp_receiver: BaseReceiver 
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_swept_center_frequency(parameters,
                                        usrp_receiver.get_spec(SpecName.API_RETUNING_LATENCY))
        validate_sample_rate_with_master_clock_rate(parameters)
    return pvalidator


def get_capture_template_fixed_center_frequency(
    usrp_receiver: BaseReceiver
) -> CaptureTemplate:
    
    capture_template = get_base_capture_template( CaptureMode.FIXED_CENTER_FREQUENCY )
    capture_template.add_ptemplate( get_base_ptemplate(PName.BANDWIDTH) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.GAIN) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.WIRE_FORMAT) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.MASTER_CLOCK_RATE) ) 

    # TODO: Delegate defaults to receiver subclasses. Currently, these are sensible defaults for the b200mini
    capture_template.set_defaults(
        (PName.BATCH_SIZE,            4.0),
        (PName.CENTER_FREQUENCY,      95800000),
        (PName.SAMPLE_RATE,           2000000),
        (PName.BANDWIDTH,             2000000),
        (PName.WINDOW_HOP,            512),
        (PName.WINDOW_SIZE,           1024),
        (PName.WINDOW_TYPE,           "blackman"),
        (PName.GAIN,                  35),
        (PName.WIRE_FORMAT,           "sc16"),
        (PName.MASTER_CLOCK_RATE,     40e6)
    )   

    capture_template.add_pconstraint(
        PName.CENTER_FREQUENCY,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec( SpecName.BANDWIDTH_LOWER_BOUND ),
                upper_bound=usrp_receiver.get_spec( SpecName.BANDWIDTH_UPPER_BOUND )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.GAIN,
        [
            Bound(
                lower_bound=0,
                upper_bound=usrp_receiver.get_spec( SpecName.GAIN_UPPER_BOUND )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.WIRE_FORMAT,
        [
            OneOf(
                usrp_receiver.get_spec( SpecName.WIRE_FORMATS )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.MASTER_CLOCK_RATE,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec( SpecName.MASTER_CLOCK_RATE_LOWER_BOUND ),
                upper_bound=usrp_receiver.get_spec( SpecName.MASTER_CLOCK_RATE_UPPER_BOUND )
            )
        ]
    )
    return capture_template


def get_capture_template_swept_center_frequency(
    usrp_receiver: BaseReceiver
) -> CaptureTemplate:

    capture_template = get_base_capture_template( CaptureMode.SWEPT_CENTER_FREQUENCY )
    capture_template.add_ptemplate( get_base_ptemplate(PName.BANDWIDTH) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.GAIN) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.WIRE_FORMAT) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.MASTER_CLOCK_RATE) ) 

    # TODO: Delegate defaults to receiver subclasses. Currently, these are sensible defaults for the b200mini
    capture_template.set_defaults(
        (PName.BATCH_SIZE,            4.0),
        (PName.MIN_FREQUENCY,         95000000),
        (PName.MAX_FREQUENCY,         105000000),
        (PName.SAMPLES_PER_STEP,      30000),
        (PName.FREQUENCY_STEP,        2000000),
        (PName.SAMPLE_RATE,           2000000),
        (PName.BANDWIDTH,             2000000),
        (PName.WINDOW_HOP,            512),
        (PName.WINDOW_SIZE,           1024),
        (PName.WINDOW_TYPE,           "blackman"),
        (PName.GAIN,                  35),
        (PName.WIRE_FORMAT,           "sc16"),
        (PName.MASTER_CLOCK_RATE,     40e6)
    )   

    capture_template.add_pconstraint(
        PName.MIN_FREQUENCY,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.MAX_FREQUENCY,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            Bound(
                lower_bound=usrp_receiver.get_spec(SpecName.BANDWIDTH_LOWER_BOUND),
                upper_bound=usrp_receiver.get_spec(SpecName.BANDWIDTH_UPPER_BOUND),
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.GAIN,
        [
            Bound(
                lower_bound=0,
                upper_bound=usrp_receiver.get_spec( SpecName.GAIN_UPPER_BOUND )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.WIRE_FORMAT,
        [
            OneOf(
                usrp_receiver.get_spec( SpecName.WIRE_FORMATS )
            )
        ]
    )
    capture_template.add_pconstraint(                                                           
        PName.MASTER_CLOCK_RATE,
        [
            Bound(                                                                              
                lower_bound=usrp_receiver.get_spec( SpecName.MASTER_CLOCK_RATE_LOWER_BOUND ),   
                upper_bound=usrp_receiver.get_spec( SpecName.MASTER_CLOCK_RATE_UPPER_BOUND )    
            )                                                                                   
        ]                                                                                       
    )                                                                                           
    return capture_template



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Type, Callable, TypeVar

from ._base import BaseBatch
from .plugins._batch_keys import BatchKey

# Map populated at runtime via the `register_batch` decorator.
batch_map: dict[BatchKey, Type[BaseBatch]] = {}

T = TypeVar('T', bound=BaseBatch)
def register_batch(
    batch_key: BatchKey
) -> Callable[[Type[T]], Type[T]]:
    """Decorator to register a `BaseBatch` subclass under a specified `BatchKey`.

    :param batch_key: The key to register the `BaseBatch` subclass under.
    :raises ValueError: If the provided `batch_key` is already registered.
    :return: A decorator that registers the `BaseBatch` subclass under the given `batch_key`.
    """
    def decorator(
        cls: Type[T]
    ) -> Type[T]:
        if batch_key in batch_map:
            raise ValueError(f"A batch with key '{batch_key}' is already registered!")
        batch_map[batch_key] = cls
        return cls
    return decorator


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from datetime import datetime

import numpy as np
import numpy.typing as npt
from astropy.io import fits
from astropy.io.fits.hdu.image import PrimaryHDU
from astropy.io.fits.hdu.table import BinTableHDU
from astropy.io.fits.hdu.hdulist import HDUList

from spectre_core.spectrograms import Spectrogram, SpectrumUnit
from ._batch_keys import BatchKey
from .._base import BaseBatch, BatchFile
from .._register import register_batch


@dataclass(frozen=True)
class _BatchExtension:
    """Supported extensions for a `CallistoBatch`.
    
    :ivar FITS: Corresponds to the `.fits` file extension.
    """
    FITS: str = "fits"
    
        
class _FitsFile(BatchFile[Spectrogram]):
    """Stores spectrogram data in the FITS format generated by the e-Callisto network."""
    def __init__(
        self, 
        parent_dir_path: str, 
        base_file_name: str
    ) -> None:
        """Initialise a `_FitsFile` instance.

        :param parent_dir_path: The parent directory for the batch.
        :param base_file_name: The batch name.
        """
        super().__init__(parent_dir_path, 
                         base_file_name, 
                         _BatchExtension.FITS)   
        
        
    def _read(
        self
    ) -> Spectrogram:
        """Parses a FITS file to generate a `Spectrogram` instance.

        Reverses the spectra along the frequency axis and converts units to linearised 
        values if necessary. Infers the spectrum type from the `BUNIT` header.

        :raises NotImplementedError: If the `BUNIT` header value represents an unsupported spectrum type.
        :return: A `Spectrogram` instance containing the parsed FITS file data.
        """
        with fits.open(self.file_path, mode='readonly') as hdulist:
            primary_hdu                = self._get_primary_hdu(hdulist)
            dynamic_spectra            = self._get_dynamic_spectra(primary_hdu)
            spectrogram_start_datetime = self._get_spectrogram_start_datetime(primary_hdu)
            bintable_hdu               = self._get_bintable_hdu(hdulist)
            times                      = self._get_times(bintable_hdu)
            frequencies                = self._get_frequencies(bintable_hdu)
            bunit                      = self._get_bunit(primary_hdu)

            # bunit is interpreted as a SpectrumUnit
            spectrum_unit = SpectrumUnit(bunit)
            if spectrum_unit == SpectrumUnit.DIGITS:
                dynamic_spectra_linearised = self._convert_units_to_linearised(dynamic_spectra)
                
                return Spectrogram(dynamic_spectra_linearised[::-1, :], # reverse the spectra along the frequency axis
                                   times, 
                                   frequencies[::-1], # sort the frequencies in ascending order
                                   self.tag, 
                                   spectrum_unit,
                                   spectrogram_start_datetime)
            else:
                raise NotImplementedError(f"SPECTRE does not currently support spectrum type with BUNITS '{spectrum_unit}'")   


    def _get_primary_hdu(
        self, hdulist: HDUList
    ) -> PrimaryHDU:
        return hdulist['PRIMARY']
    

    def _get_bintable_hdu(
        self, 
        hdulist: HDUList
    ) -> BinTableHDU:
        return hdulist[1]
    
    
    def _get_dynamic_spectra(
        self, 
        primary_hdu: PrimaryHDU
    ) -> npt.NDArray[np.float32]:
        return primary_hdu.data.astype(np.float32)


    def _get_spectrogram_start_datetime(
        self, 
        primary_hdu: PrimaryHDU
    ) -> datetime:
        date_obs = primary_hdu.header['DATE-OBS']
        time_obs = primary_hdu.header['TIME-OBS']
        return datetime.strptime(f"{date_obs}T{time_obs}", "%Y/%m/%dT%H:%M:%S.%f")


    def _get_bunit(
        self, 
        primary_hdu: PrimaryHDU
    ) -> str:
        return primary_hdu.header['BUNIT']


    def _convert_units_to_linearised(
        self, 
        raw_digits: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Converts spectrogram data from raw digit values to linearised units.

        Applies a transformation based on ADC specifications to convert raw values 
        to dB and then to linearised units.

        :param dynamic_spectra: Raw dynamic spectra in digit values.
        :return: The dynamic spectra with linearised units.
        """
        # conversion as per ADC specs [see email from C. Monstein]
        dB = (raw_digits / 255) * (2500 / 25)
        return 10 ** (dB / 10)
    
    
    def _get_times(
        self, 
        bintable_hdu: BinTableHDU
    ) -> npt.NDArray[np.float32]:
        """Extracts the elapsed times for each spectrum in seconds, with the first spectrum set to t=0
        by convention.
        """
        return bintable_hdu.data['TIME'][0] # already in seconds


    def _get_frequencies(
        self,
        bintable_hdu: BinTableHDU
    ) -> npt.NDArray[np.float32]:
        """Extracts the frequencies for each spectral component."""
        frequencies_MHz = bintable_hdu.data['FREQUENCY'][0]
        return frequencies_MHz * 1e6 # convert to Hz
    
  
@register_batch(BatchKey.CALLISTO)
class CallistoBatch(BaseBatch):
    """A batch of data generated by the e-Callisto network.
    
    Supports the following file extensions:
    - `.fits` (via the `spectrogram_file` attribute)
    """
    def __init__(
        self,
        start_time: str,
        tag: str
    ) -> None:
        """Initialise a `CallistoBatch` instance.

        :param start_time: The start time of the batch.
        :param tag: The batch name tag.
        """
        super().__init__(start_time, tag) 
        self._fits_file = _FitsFile(self.parent_dir_path, self.name)
        
        # add files formally to the batch
        self.add_file( self.spectrogram_file )
    
    
    @property
    def spectrogram_file(
        self
    ) -> _FitsFile:
        """The batch file corresponding to the `.fits` extension."""
        return self._fits_file



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Optional, Any
from functools import cached_property

from spectre_core._file_io import JsonHandler
from spectre_core.config import get_configs_dir_path
from spectre_core.exceptions import InvalidTagError
from ._ptemplates import PName
from ._parameters import (
    Parameter, 
    Parameters,
    make_parameters
)

@dataclass(frozen=True)
class _CaptureConfigKey:
    """Defined JSON keys for capture configs.
    
    :ivar RECEIVER_NAME: The name of the receiver used for capture.
    :ivar RECEIVER_MODE: The operating mode for the receiver to be used for capture.
    :ivar PARAMETERS: The user configured parameters given to the receiver at the time of capture.
    """
    RECEIVER_NAME = "receiver_name"
    RECEIVER_MODE = "receiver_mode"
    PARAMETERS    = "parameters"


class CaptureConfig(JsonHandler):
    """Simple IO interface for a capture configs under a particular tag."""
    def __init__(
        self,
        tag: str
    ) -> None:
        """Initialise an instance of `CaptureConfig`.

        :param tag: The tag identifier for the capture config.
        """
        self._validate_tag(tag)
        self._tag = tag
        super().__init__(get_configs_dir_path(),
                         f"capture_{tag}")
        
    @property
    def tag(
        self
    ) -> str:
        """The tag identifier for the capture config."""
        return self._tag


    def _validate_tag(
        self, 
        tag: str
    ) -> None:
        """Validate the tag of the capture config.
        
        Some substrings are reserved for third-party spectrogram data.
        """
        if "_" in tag:
            raise InvalidTagError(f"Tags cannot contain an underscore. Received {tag}")
        
        if "callisto" in tag:
            raise ValueError(f"The substring `callisto` is not allowed in a capture config tag.")
    

    @property
    def receiver_name(
        self
    ) -> str:
        """The name of the receiver to be used for capture."""
        d = self.read(cache=True)
        return d[_CaptureConfigKey.RECEIVER_NAME]
    

    @property
    def receiver_mode(
        self
    ) -> str:
        """The operating mode for the receiver to be used for capture."""
        d = self.read(cache=True)
        return d[_CaptureConfigKey.RECEIVER_MODE]
    

    @cached_property
    def parameters(
        self
    ) -> Parameters:
        """The user-configured parameters provided to the receiver at the time of capture."""
        d = self.read(cache=True)
        return make_parameters( d[_CaptureConfigKey.PARAMETERS] )


    def get_parameter(
        self, 
        name: PName
    ) -> Parameter:
        """Get a parameter stored by the capture config.

        :param name: The name of the parameter.
        :return: A `Parameter` instance with `name` and `value` retrieved from the capture
                configuration file.
        """
        return self.parameters.get_parameter(name)
    

    def get_parameter_value(
        self,
        name: PName
    ) -> Optional[Any]:
        """Get the value of a parameter stored by the capture config.
        
        For static typing, should be `cast` after return according to the corresponding `PTemplate`.

        :param name: The name of the parameter.
        :return: The value of the parameter corresponding to `name`. If the JSON value is
        `null`, this method will return `None`.
        """
        return self.parameters.get_parameter_value(name)


    def save_parameters(
        self,
        receiver_name: str,
        receiver_mode: str,
        parameters: Parameters,
        force: bool = False
    ):
        """Write the input parameters to a capture config.

        :param receiver_name: The name of the receiver to be used for capture.
        :param receiver_mode: The operating mode for the receiver to be used for capture.
        :param parameters: The user-configured parameters provided to the receiver at the time of capture.
        :param force: If true, force the write if the file already exists in the file system. Defaults to False.
        """
        d = {
            _CaptureConfigKey.RECEIVER_MODE: receiver_mode,
            _CaptureConfigKey.RECEIVER_NAME: receiver_name,
            _CaptureConfigKey.PARAMETERS   : parameters.to_dict()
        }
        self.save(d, force=force)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass

from ._receiver_names import ReceiverName
from ._usrp import (
        get_pvalidator_fixed_center_frequency, get_pvalidator_swept_center_frequency,
        get_capture_template_fixed_center_frequency, get_capture_template_swept_center_frequency
)
from .gr._usrp import CaptureMethod
from .._spec_names import SpecName
from .._base import BaseReceiver
from .._register import register_receiver 

@dataclass(frozen=True)
class Mode:
    """An operating mode for the `B200mini` receiver."""
    FIXED_CENTER_FREQUENCY = "fixed_center_frequency"
    SWEPT_CENTER_FREQUENCY = "swept_center_frequency"


@register_receiver(ReceiverName.B200MINI)
class B200mini(BaseReceiver):
    """Receiver implementation for the USRP B200mini (https://www.ettus.com/all-products/usrp-b200mini/)"""
    def _add_specs(
        self
    ) -> None:
        self.add_spec( SpecName.SAMPLE_RATE_LOWER_BOUND      , 200e3 )
        self.add_spec( SpecName.SAMPLE_RATE_UPPER_BOUND      , 56e6  )
        self.add_spec( SpecName.FREQUENCY_LOWER_BOUND        , 70e6  )
        self.add_spec( SpecName.FREQUENCY_UPPER_BOUND        , 6e9   )
        self.add_spec( SpecName.BANDWIDTH_LOWER_BOUND        , 200e3 ) 
        self.add_spec( SpecName.BANDWIDTH_UPPER_BOUND        , 56e6  )
        self.add_spec( SpecName.GAIN_UPPER_BOUND             , 76    )
        self.add_spec( SpecName.WIRE_FORMATS                 , ["sc8", "sc12", "sc16"])
        self.add_spec( SpecName.MASTER_CLOCK_RATE_LOWER_BOUND, 5e6)
        self.add_spec( SpecName.MASTER_CLOCK_RATE_UPPER_BOUND, 61.44e6)
        self.add_spec( SpecName.API_RETUNING_LATENCY         , 1e-5  ) # TODO: This is a ballpark, pending empirical testing 

   
    def _add_capture_methods(
        self
    ) -> None:
       self.add_capture_method(Mode.FIXED_CENTER_FREQUENCY,
                               CaptureMethod.fixed_center_frequency)
       self.add_capture_method(Mode.SWEPT_CENTER_FREQUENCY,
                               CaptureMethod.swept_center_frequency)

   
    def _add_capture_templates(
        self
    ) -> None:
        self.add_capture_template(Mode.FIXED_CENTER_FREQUENCY,
                                  get_capture_template_fixed_center_frequency(self))
        self.add_capture_template(Mode.SWEPT_CENTER_FREQUENCY,
                                  get_capture_template_swept_center_frequency(self))


    def _add_pvalidators(
        self
    ) -> None:
        self.add_pvalidator(Mode.FIXED_CENTER_FREQUENCY,
                            get_pvalidator_fixed_center_frequency(self))
        self.add_pvalidator(Mode.SWEPT_CENTER_FREQUENCY,
                            get_pvalidator_swept_center_frequency(self))

    
 

















# 
# USRP top blocks
#

from functools import partial
from dataclasses import dataclass
import time

from logging import getLogger
_LOGGER = getLogger(__name__)

from spectre_core.capture_configs import Parameters, PName
from spectre_core.config import get_batches_dir_path
from ._base import capture, spectre_top_block


class _fixed_center_frequency(spectre_top_block):
    def flowgraph(
        self,
        tag: str,
        parameters: Parameters
    ) -> None:
        # OOT moudle inline imports
        from gnuradio import spectre
        from gnuradio import uhd

        # Unpack capture config parameters 
        sample_rate       = parameters.get_parameter_value(PName.SAMPLE_RATE)
        gain              = parameters.get_parameter_value(PName.GAIN)
        center_freq       = parameters.get_parameter_value(PName.CENTER_FREQUENCY)
        master_clock_rate = parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
        wire_format       = parameters.get_parameter_value(PName.WIRE_FORMAT)
        batch_size        = parameters.get_parameter_value(PName.BATCH_SIZE)
        bandwidth         = parameters.get_parameter_value(PName.BANDWIDTH) 

        # Blocks
        master_clock_rate = f"master_clock_rate={master_clock_rate}" 
        self.uhd_usrp_source_0 = uhd.usrp_source(
            ",".join(("", '', master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=wire_format,
                args='',
                channels=[0],
            ),
        )
        self.uhd_usrp_source_0.set_samp_rate(sample_rate)
        self.uhd_usrp_source_0.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)

        self.uhd_usrp_source_0.set_center_freq(center_freq, 0)
        self.uhd_usrp_source_0.set_antenna("RX2", 0)
        self.uhd_usrp_source_0.set_bandwidth(bandwidth, 0)
        self.uhd_usrp_source_0.set_rx_agc(False, 0)
        self.uhd_usrp_source_0.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source_0.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source_0.set_gain(gain, 0)
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_batches_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     sample_rate, False, 
                                                                     'rx_freq', 
                                                                     0)


        # Connections
        self.connect((self.uhd_usrp_source_0, 0), (self.spectre_batched_file_sink_0, 0))


class _swept_center_frequency(spectre_top_block):
    def flowgraph(
        self,
        tag: str,
        parameters: Parameters
    ) -> None:
        # OOT module inline imports
        from gnuradio import spectre
        from gnuradio import uhd

        # Unpack capture config parameters
        sample_rate       = parameters.get_parameter_value(PName.SAMPLE_RATE)
        bandwidth         = parameters.get_parameter_value(PName.BANDWIDTH)
        min_frequency     = parameters.get_parameter_value(PName.MIN_FREQUENCY)
        max_frequency     = parameters.get_parameter_value(PName.MAX_FREQUENCY)
        frequency_step    = parameters.get_parameter_value(PName.FREQUENCY_STEP)
        samples_per_step  = parameters.get_parameter_value(PName.SAMPLES_PER_STEP)
        master_clock_rate = parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
        master_clock_rate = master_clock_rate = parameters.get_parameter_value(PName.MASTER_CLOCK_RATE)
        wire_format       = parameters.get_parameter_value(PName.WIRE_FORMAT)
        gain              = parameters.get_parameter_value(PName.GAIN)
        batch_size        = parameters.get_parameter_value(PName.BATCH_SIZE)

        # Blocks
        _LOGGER.warning(f"USRP frequency sweep modes will not work as expected until a known bug is fixed in the USRP source block. "
                        f"Please refer to this GitHub issue for more information: https://github.com/gnuradio/gnuradio/issues/7725")
        master_clock_rate = f"master_clock_rate={master_clock_rate}" 
        self.uhd_usrp_source_0 = uhd.usrp_source(
            ",".join(("", '', master_clock_rate)),
            uhd.stream_args(
                cpu_format="fc32",
                otw_format=wire_format,
                args='',
                channels=[0],
            ),
        )
        self.uhd_usrp_source_0.set_samp_rate(sample_rate)
        self.uhd_usrp_source_0.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source_0.set_center_freq(min_frequency, 0)
        self.uhd_usrp_source_0.set_antenna("RX2", 0)
        self.uhd_usrp_source_0.set_bandwidth(bandwidth, 0)
        self.uhd_usrp_source_0.set_rx_agc(False, 0)
        self.uhd_usrp_source_0.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source_0.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source_0.set_gain(gain, 0)

        self.spectre_sweep_driver_0 = spectre.sweep_driver(min_frequency, 
                                                           max_frequency, 
                                                           frequency_step, 
                                                           sample_rate, 
                                                           samples_per_step, 
                                                           'freq')
       
        self.spectre_batched_file_sink_0 = spectre.batched_file_sink(get_batches_dir_path(), 
                                                                     tag, 
                                                                     batch_size, 
                                                                     sample_rate,
                                                                     True, 
                                                                     'rx_freq', 
                                                                     min_frequency)

        # Connections
        self.msg_connect((self.spectre_sweep_driver_0, 'freq'), (self.uhd_usrp_source_0, 'command'))
        self.connect((self.uhd_usrp_source_0, 0), (self.spectre_batched_file_sink_0, 0))
        self.connect((self.uhd_usrp_source_0, 0), (self.spectre_sweep_driver_0, 0))


@dataclass(frozen=True)
class CaptureMethod:
    fixed_center_frequency = partial(capture, top_block_cls=_fixed_center_frequency)
    swept_center_frequency = partial(capture, top_block_cls=_swept_center_frequency, max_noutput_items=1024)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import json
from abc import ABC, abstractmethod
from typing import Any, Optional, TypeVar, Generic

T = TypeVar('T')

class BaseFileHandler(ABC, Generic[T]):
    """
    Base class for handling file operations.

    When subclassing, specify the return type of `_read` 
    using `Generic[T]`. 

    Example:
    .. code-block:: python

        from typing import Any, Generic

        class JsonHandler(BaseFileHandler[dict[str, Any]]):
            def _read(self) -> dict[str, Any]:
                # Implementation here
                ...
    """
    def __init__(
        self, 
        parent_dir_path: str, 
        base_file_name: str, 
        extension: Optional[str] = None
    ) -> None:
        """Initialise a `BaseFileHandler` instance.

        :param parent_dir_path: The directory path where the file is located (absolute or relative).
        :param base_file_name: The name of the file without its extension.
        :param extension: The file extension (without the dot), defaults to None
        """
        self._data_cache: Optional[T] = None
        
        self._parent_dir_path = parent_dir_path
        self._base_file_name  = base_file_name
        
        if extension == "":
            extension = None
        self._extension = extension

     
    @abstractmethod
    def _read(
        self
    ) -> T:
        """The grunt work to return the file contents.

        :return: The file contents.
        """
     
    
    @property
    def parent_dir_path(
        self
    ) -> str:
        """Return the parent directory path for the file."""
        return self._parent_dir_path
    

    @property
    def base_file_name(
        self
    ) -> str:
        """Return the file name, stripped of the file extension."""
        return self._base_file_name
    

    @property
    def extension(
        self
    ) -> Optional[str]:
        """Return the file path suffix, excluding the dot."""
        return self._extension
    

    @property
    def file_name(
        self
    ) -> str:
        """Generate the file name based on the base name and extension.

        :return: The file name with the extension (including the dot), or the base name if no extension is set.
        """
        return self._base_file_name if (self._extension is None) else f"{self._base_file_name}.{self._extension}"
    

    @property
    def file_path(
        self
    ) -> str:
        """The absolute or relative file path as defined by the parent directory path,
        base file name and extension."""
        return os.path.join(self._parent_dir_path, self.file_name)
    
    
    @property
    def exists(
        self
    ) -> bool:
        """Check if the file exists in the filesystem."""
        return os.path.exists(self.file_path) 


    def read(
        self,
        cache: bool = True
    ) -> T:
        """Read the file contents.
        
        :param cache: If False, bypasses the cache and reads the file directly on each `read` call, defaults to True
        :return: The file contents.
        """
        # if the user has specified to ignore the cache, simply read the file.
        if not cache:
            return self._read()
        
        # otherwise make use of the cache
        if self._data_cache is None:
            self._data_cache = self._read()
        return self._data_cache
    
    
    def make_parent_dir_path(
        self
    ) -> None:
        """Make the parent directory path of the file. No error is raised if the target 
        directory already exists.
        """
        os.makedirs(self.parent_dir_path, exist_ok=True) 
    

    def delete(
        self,
        ignore_if_missing: bool = False
    ) -> None:
        """Delete the file from the filesystem.

        :param ignore_if_missing: If True, skips deletion if the file does not exist, defaults to False
        :raises FileNotFoundError: If the file is missing and `ignore_if_missing` is False.
        """
        if not self.exists and not ignore_if_missing:
            raise FileNotFoundError(f"{self.file_name} does not exist, and so cannot be deleted")
        else:
            os.remove(self.file_path)
    

    def cat(
        self
    ) -> None:
        """Display the file contents on the standard output."""
        print(self.read())


class JsonHandler(BaseFileHandler[dict[str, Any]]):
    """File handler for JSON formatted files.
    
    We assume that the files are of the form
    {
        "foo": <JSON compatable structure>
        ... and so on.
    }

    """
    def __init__(
        self, 
        parent_dir_path: str, 
        base_file_name: str,
        extension: str = "json"
    ) -> None:
        super().__init__(parent_dir_path, 
                         base_file_name, 
                         extension)
    
    
    def _read(
        self
    ) -> dict[str, Any]:
        with open(self.file_path, 'r') as f:
            return json.load(f)
        
        
    def save(
        self, 
        d: dict[str, Any], 
        force: bool = False
    ) -> None:
        """Save the input dictionary to file in the JSON file format.

        :param d: The dictionary to save.
        :param force: If True, overwrites the file if it already exists, defaults to False
        :raises FileExistsError: If the file exists and `force` is False.
        """
        self.make_parent_dir_path()

        if self.exists:
            if force:
                pass
            else:
                raise FileExistsError((f"{self.file_name} already exists, write has been abandoned. "
                                       f"You can override this functionality with `force`"))

        with open(self.file_path, 'w') as file:
                json.dump(d, file, indent=4)

    

class TextHandler(BaseFileHandler[str]):
    """File handler for text formatted files."""
    def __init__(
        self, 
        parent_dir_path: str,
        base_file_name: str, 
        extension: str = "txt"
    ) -> None:
        super().__init__(parent_dir_path,
                         base_file_name,
                         extension)
    
    
    def _read(
        self
    ) -> str:
        with open(self.file_path, 'r') as f:
            return f.read()


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from functools import wraps
import time
from typing import Callable, TypeVar, ParamSpec
import multiprocessing

from spectre_core.logs import configure_root_logger, log_call, ProcessType
from spectre_core.capture_configs import CaptureConfig
from spectre_core.receivers import get_receiver, ReceiverName
from spectre_core.post_processing import start_post_processor


def _make_daemon_process(
    name: str, 
    target: Callable[[], None]
) -> multiprocessing.Process:
    """
    Creates and returns a daemon `multiprocessing.Process` instance.

    :param name: The name to assign to the process.
    :param target: The function to execute in the process.
    :return: A `multiprocessing.Process` instance configured as a daemon.
    """
    return multiprocessing.Process(target=target,
                                   name=name,
                                   daemon=True)


class Worker:
    """A lightweight wrapper for a `multiprocessing.Process` daemon.
    
    Provides a very simple API to start, and restart a multiprocessing process.
    """
    def __init__(
        self,
        name: str,
        target: Callable[[], None]
    ) -> None:
        """Initialise a `Worker` instance.

        :param name: The name assigned to the process.
        :param target: The callable to be executed by the worker process.
        """
        self._name = name
        self._target = target
        self._process = _make_daemon_process(name, target)


    @property
    def name(
        self
    ) -> str:
        """Get the name of the worker process.

        :return: The name of the multiprocessing process.
        """
        return self._process.name
    
    
    @property
    def process(
        self
    ) -> multiprocessing.Process:
        """Access the underlying multiprocessing process.

        :return: The wrapped `multiprocessing.Process` instance.
        """
        return self._process

    
    def start(
        self
    ) -> None:
        """Start the worker process.

        This method runs the `target` in the background as a daemon.
        """
        self._process.start()


    def restart(
        self
    ) -> None:
        """Restart the worker process.

        Terminates the existing process if it is alive and then starts a new process
        after a brief pause. 
        """
        _LOGGER.info(f"Restarting {self.name} worker")
        if self._process.is_alive():
            # forcibly stop if it is still alive
            self._process.terminate()
            self._process.join()
        # a moment of respite
        time.sleep(1)
        # make a new process, as we can't start the same process again.
        self._process = _make_daemon_process(self._name, self._target)
        self.start()


P = ParamSpec("P")
T = TypeVar("T", bound=Callable[..., None])
def make_worker(
    name: str
) -> Callable[[Callable[P, None]], Callable[P, Worker]]:
    """
    Turns a function into a worker.

    This decorator wraps a function, allowing it to run in a separate process
    managed by a `Worker` object. Use it to easily create long-running or
    isolated tasks without directly handling multiprocessing.

    :param name: A human-readable name for the worker process.
    :return: A decorator that creates a `Worker` to run the function in its own process.
    """

    def decorator(
        func: Callable[P, None]
    ) -> Callable[P, Worker]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Worker:
            # Worker target funcs must have no arguments
            def target():
                configure_root_logger(ProcessType.WORKER)
                func(*args, **kwargs)
            return Worker(name, target)
        return wrapper
    return decorator


@make_worker("capture")
@log_call
def do_capture(
    tag: str,
) -> None:
    """Start capturing data from an SDR in real time.

    :param tag: The capture config tag.
    """
    _LOGGER.info((f"Reading capture config with tag '{tag}'"))

    # load the receiver and mode from the capture config file
    capture_config = CaptureConfig(tag)

    _LOGGER.info((f"Starting capture with the receiver '{capture_config.receiver_name}' "
                  f"operating in mode '{capture_config.receiver_mode}' "
                  f"with tag '{tag}'"))

    name = ReceiverName( capture_config.receiver_name )
    receiver = get_receiver(name,
                            capture_config.receiver_mode)
    receiver.start_capture(tag)


@make_worker("post_processing")
@log_call
def do_post_processing(
    tag: str,
) -> None:
    """Start post processing SDR data into spectrograms in real time.

    :param tag: The capture config tag.
    """
    _LOGGER.info(f"Starting post processor with tag '{tag}'")
    start_post_processor(tag)


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later


"""General `spectre` package configurations."""

from ._paths import (
    get_spectre_data_dir_path, get_batches_dir_path, get_configs_dir_path, 
    get_logs_dir_path, trim_spectre_data_dir_path
)
from ._time_formats import (
    TimeFormat
)

__all__ = [
    "get_spectre_data_dir_path", "get_batches_dir_path", "get_configs_dir_path", "get_logs_dir_path",
    "TimeFormat", "trim_spectre_data_dir_path"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

from typing import Optional, cast
from abc import ABC, abstractmethod
from scipy.signal import ShortTimeFFT, get_window

from watchdog.events import FileSystemEventHandler, FileSystemEvent

from spectre_core.capture_configs import CaptureConfig, PName
from spectre_core.spectrograms import Spectrogram, join_spectrograms


def make_sft_instance(
    capture_config: CaptureConfig
) -> ShortTimeFFT:
    """Extract window parameters from the input capture config and create an instance
    of `ShortTimeFFT` from `scipy.signal`.

    :param capture_config: The capture config storing the parameters.
    :return: An instance of `ShortTimeFFT` consistent with the window parameters 
    in the capture config.
    """
    window_type   = cast(str, capture_config.get_parameter_value(PName.WINDOW_TYPE))
    sample_rate   = cast(int, capture_config.get_parameter_value(PName.SAMPLE_RATE))
    window_hop    = cast(int, capture_config.get_parameter_value(PName.WINDOW_HOP))
    window_size   = cast(int, capture_config.get_parameter_value(PName.WINDOW_SIZE))
    window = get_window(window_type, 
                        window_size)
    return ShortTimeFFT(window, 
                        window_hop,
                        sample_rate, 
                        fft_mode = "centered")


class BaseEventHandler(ABC, FileSystemEventHandler):
    """An abstract base class for event-driven file post-processing."""
    def __init__(
        self, 
        tag: str
    ) -> None:
        """Initialise a `BaseEventHandler` instance.

        :param tag: The tag of the capture config used to capture the data.
        """
        self._tag = tag

        # load the capture config corresponding to the input tag
        self._capture_config = CaptureConfig(tag)
        
        # store the next file to be processed (specifically, the absolute file path of the file)
        self._queued_file: Optional[str] = None

        # optionally store batched spectrograms as they are created into a cache
        # this can be flushed periodically to file as required.
        self._cached_spectrogram: Optional[Spectrogram] = None


    @abstractmethod
    def process(
        self, 
        absolute_file_path: str
    ) -> None:
        """
        Process a batch file at the given file path.

        :param absolute_file_path: The absolute path to the batch file to be processed.
        """


    def on_created(
        self, 
        event: FileSystemEvent
    ) -> None:
        """Process a newly created batch file, only once the next batch is created.
        
        Since we assume that the batches are non-overlapping in time, this guarantees
        we avoid post processing a file while it is being written to. Files are processed
        sequentially, in the order they are created.

        :param event: The file system event containing the file details.
        """
        # the `src_path`` attribute holds the absolute path of the freshly closed file
        absolute_file_path = event.src_path
        
        # only 'notice' a file if it ends with the appropriate extension as defined in the capture config
        watch_extension = cast(str, self._capture_config.get_parameter_value(PName.WATCH_EXTENSION))
        
        if absolute_file_path.endswith( watch_extension ):
            _LOGGER.info(f"Noticed {absolute_file_path}")
            # If there exists a queued file, try and process it
            if self._queued_file is not None:
                try:
                    self.process(self._queued_file)
                except Exception:
                    _LOGGER.error(f"An error has occured while processing {self._queued_file}",
                                  exc_info=True)
                     # flush any internally stored spectrogram on error to avoid lost data
                    self._flush_cache()
                    # re-raise the exception to the main thread
                    raise
            
            # Queue the current file for processing next
            _LOGGER.info(f"Queueing {absolute_file_path} for post processing")
            self._queued_file = absolute_file_path


    def _cache_spectrogram(
        self, 
        spectrogram: Spectrogram
    ) -> None:
        """Cache the input spectrogram by storing it in the `_cached_spectrogram` attribute.
        
        If the time range of the cached spectrogram exceeds that as specified in the capture config
        `PName.TIME_RANGE` parameter, the spectrogram in the cache is flushed to file. If `PName.TIME_RANGE`
        is nulled, the cache is flushed immediately.

        :param spectrogram: The spectrogram to store in the cache.
        """
        _LOGGER.info("Joining spectrogram")

        if self._cached_spectrogram is None:
            self._cached_spectrogram = spectrogram
        else:
            self._cached_spectrogram = join_spectrograms([self._cached_spectrogram, spectrogram])
  
        time_range = self._capture_config.get_parameter_value(PName.TIME_RANGE) or 0.0
        if self._cached_spectrogram.time_range >= cast(float, time_range):
            self._flush_cache()
    

    def _flush_cache(
        self
    ) -> None:
        """Flush the cached spectrogram to file."""
        if self._cached_spectrogram:
            _LOGGER.info(f"Flushing spectrogram to file with start time "
                         f"'{self._cached_spectrogram.format_start_time()}'")
            self._cached_spectrogram.save()
            _LOGGER.info("Flush successful, resetting spectrogram cache")
            self._cached_spectrogram = None # reset the cache


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Optional, Literal
from enum import Enum

import numpy.typing as npt
import numpy as np
import matplotlib.dates as mdates
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from spectre_core.spectrograms import Spectrogram, TimeType
from ._format import PanelFormat
from ._panel_names import PanelName


class XAxisType(Enum):
    """The x-axis type for a panel.
    
    Axes are shared in a stack between panels with common `XAxisType`.
    
    :ivar TIME: The xaxis has units of time.
    :ivar FREQUENCY: The xaxis has units of frequency.
    """
    TIME      = "time"
    FREQUENCY = "frequency"


class BasePanel(ABC):
    """Abstract base class for a panel used to visualise spectrogram data.

    `BasePanel` instances are designed to be part of a `PanelStack`, where multiple
    panels contribute to a composite plot. Subclasses must implement methods to define
    how the panel is drawn and annotated, and specify its x-axis type.
    """
    def __init__(
        self, 
        name: PanelName,
        spectrogram: Spectrogram
    ) -> None:
        """Initialize an instance of `BasePanel`.

        :param name: The name of the panel.
        :param spectrogram: The spectrogram being visualised.
        """
        self._name = name
        self._spectrogram = spectrogram

        # internal attributes set by `PanelStack` during stacking.
        self._panel_format: Optional[PanelFormat] = None
        self._time_type   : Optional[TimeType]    = None
        self._ax          : Optional[Axes]        = None
        self._fig         : Optional[Figure]      = None
        self._identifier  : Optional[str]         = None


    @abstractmethod
    def draw(
        self
    ) -> None:
        """Modify the `ax` attribute to draw the panel contents."""


    @abstractmethod
    def annotate_xaxis(
        self
    ) -> None:
        """Modify the `ax` attribute to annotate the x-axis of the panel."""


    @abstractmethod
    def annotate_yaxis(
        self
    ) -> None:
        """Modify the `ax` attribute to annotate the y-axis of the panel."""


    @property
    @abstractmethod
    def xaxis_type(
        self
    ) -> XAxisType:
     """Specify the x-axis type for the panel."""


    @property
    def spectrogram(
        self
    ) -> Spectrogram:
        """The spectrogram being visualised on this panel."""
        return self._spectrogram
    

    @property
    def tag(
        self
    ) -> str:
        """The tag of the spectrogram being visualised."""
        return self._spectrogram.tag
    

    @property
    def time_type(
        self
    ) -> TimeType:
        """The time type of the spectrogram.

        :raises ValueError: If the `time_type` has not been set.
        """
        if self._time_type is None:
            raise ValueError(f"`time_type` for the panel '{self.name}' must be set.")
        return self._time_type
    

    @time_type.setter
    def time_type(
        self, 
        value: TimeType
    ) -> None:
        """Set the `TimeType` for the spectrogram.

        This controls how time is represented and annotated on the panel.

        :param value: The `TimeType` to assign to the spectrogram.
        """
        self._time_type = value
    

    @property
    def name(
        self
    ) -> PanelName:
        """The name of the panel."""
        return self._name
    
    
    @property
    def panel_format(
        self
    ) -> PanelFormat:
        """Retrieve the panel format, which controls the style of the panel.

        :raises ValueError: If the `panel_format` has not been set.
        """
        if self._panel_format is None:
            raise ValueError(f"`panel_format` for the panel '{self.name}' must be set.")
        return self._panel_format


    @panel_format.setter
    def panel_format(
        self, 
        value: PanelFormat
    ) -> None:
        """Set the panel format to control the style of the panel.
        
        :param value: The `PanelFormat` to assign to the panel.
        """
        self._panel_format = value


    @property
    def ax(
        self
    ) -> Axes:
        """The `Axes` object bound to this panel.

        :raises AttributeError: If the `Axes` object has not been set.
        """
        if self._ax is None:
            raise AttributeError(f"`ax` must be set for the panel `{self.name}`")
        return self._ax
    

    @ax.setter
    def ax(
        self, 
        value: Axes
    ) -> None:
        """Assign a Matplotlib `Axes` object to this panel.

        This `Axes` will be used for drawing and annotations.

        :param value: The Matplotlib `Axes` to assign to the panel.
        """
        self._ax = value


    @property
    def fig(
        self
    ) -> Figure:
        """
        The `Figure` object bound to this panel.

        :raises AttributeError: If the `Figure` object has not been set.
        """
        if self._fig is None:
            raise AttributeError(f"`fig` must be set for the panel `{self.name}`")
        return self._fig
    

    @fig.setter
    def fig(
        self, 
        value: Figure
    ) -> None:
        """
        Assign a Matplotlib `Figure` object to this panel.

        This `Figure` is shared across all panels in the `PanelStack`.

        :param value: The Matplotlib `Figure` to assign to the panel.
        """
        self._fig = value
    
    
    @property
    def identifier(
        self
    ) -> Optional[str]:
        """Optional identifier for the panel.

        This identifier can be used to distinguish panels or aid in superimposing 
        panels in a stack.
        """
        return self._identifier
    
    
    @identifier.setter
    def identifier(
        self, 
        value: str
    ) -> None:
        """Set the optional identifier for the panel.
        
        This can be used to distinguish panels or aid in superimposing panels.
        """
        self._identifier = value
    

    def hide_xaxis_labels(
        self
    ) -> None:
        """Hide the x-axis labels for this panel."""
        self.ax.tick_params(axis='x', labelbottom=False)


    def hide_yaxis_labels(
        self
    ) -> None:
        """Hide the y-axis labels for this panel."""
        self.ax.tick_params(axis='y', labelbottom=False)
    

class BaseTimeSeriesPanel(BasePanel):
    """
    Abstract subclass of `BasePanel` designed for visualising time series data.

    Subclasses must implement any remaining abstract methods from `BasePanel`.
    """   
    @property
    def xaxis_type(
        self
    ) -> Literal[XAxisType.TIME]:
        return XAxisType.TIME
    
    
    @property
    def times(
        self
    ) -> npt.NDArray[np.float32 | np.datetime64]:
        """The times assigned to each spectrum according to the `TimeType`."""
        return self.spectrogram.times if self.time_type == TimeType.RELATIVE else self.spectrogram.datetimes
    

    def annotate_xaxis(
        self
    ) -> None:
        """Annotate the x-axis according to the specified `TimeType`."""
        if self.time_type == TimeType.RELATIVE:
            self.ax.set_xlabel('Time [s]')
        else:
            self.ax.set_xlabel('Time [UTC]')
            self.ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M:%S'))



class BaseSpectrumPanel(BasePanel):
    """An abstract subclass of `BasePanel` tailored for visualising spectrum data.
    
    Subclasses must implement any remaining abstract methods as described by `BasePanel`.
    """   
    @property
    def xaxis_type(
        self
    ) -> Literal[XAxisType.FREQUENCY]:
        return XAxisType.FREQUENCY
    
    
    @property
    def frequencies(
        self
    ) -> npt.NDArray[np.float32]:
        """The physical frequencies assigned to each spectral component."""
        return self._spectrogram.frequencies


    def annotate_xaxis(
        self
    ) -> None:
        """Annotate the x-axis assuming frequency in units of Hz."""
        self.ax.set_xlabel('Frequency [Hz]')


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from math import floor
from typing import Optional, cast

from scipy.signal import get_window

from ._parameters import Parameters
from ._pnames import PName

# ----------------------------------------------------------------------- # 
# Throughout this module, repeated calls to `cast` will be seen on using the
# `get_parameter_value` method. This is purely to signify intent and keep
# the static type checkers happy.
# 
# This is okay, as whenever the validators are called, the parameter values 
# have already been individually casted and constrained according to the 
# parameter templates. There is negligible runtime impact. Though a bit
# of a pain for maintenance.
# ----------------------------------------------------------------------- # 


def validate_window(
        parameters: Parameters
) -> None:
    """Ensure that the capture config describes a valid window.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the window interval is greater than the batch size.
    :raises ValueError: If the specified window type cannot be fetched using 
    the SciPy `get_window` function.
    """
    window_size = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
    batch_size  = cast(int, parameters.get_parameter_value(PName.BATCH_SIZE))
    window_type = cast(str, parameters.get_parameter_value(PName.WINDOW_TYPE))
    
    window_interval = window_size*(1 / sample_rate)
    if window_interval > batch_size:
        raise ValueError((f"The windowing interval must be strictly less than the batch size. "
                          f"Computed the windowing interval to be {window_interval} [s], "
                          f"but the batch size is {batch_size} [s]"))
    
    try:
        _ = get_window(window_type, window_size)
    except Exception as e:
        raise ValueError((f"An error has occurred while validating the window. "
                          f"Got {str(e)}"))
    

def validate_nyquist_criterion(
        parameters: Parameters
) -> None:
    """Ensure that the Nyquist criterion is satisfied.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the sample rate is less than the bandwidth.
    """
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))
    bandwidth   = cast(float, parameters.get_parameter_value(PName.BANDWIDTH))

    if sample_rate < bandwidth:
        raise ValueError((f"Nyquist criterion has not been satisfied. "
                          f"Sample rate must be greater than or equal to the bandwidth. "
                          f"Got sample rate {sample_rate} [Hz], and bandwidth {bandwidth} [Hz]"))
    

def _compute_num_steps_per_sweep(min_freq: float, 
                                 max_freq: float,
                                 freq_step: float) -> int:
    """Compute the number of steps in one frequency sweep.

    The center frequency starts at `min_freq` and increments in steps of `freq_step` 
    until the next step would exceed `max_freq`.

    :param min_freq: The minimum frequency of the sweep.
    :param max_freq: The maximum frequency of the sweep.
    :param freq_step: The frequency step size.
    :return: The number of steps in one frequency sweep.
    """
    return floor((max_freq - min_freq) / freq_step)


def validate_num_steps_per_sweep(
        parameters: Parameters
) -> None:
    """Ensure that there are at least two steps in frequency per sweep.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the number of steps per sweep is less than or equal to one.
    """
    min_freq    = cast(float, parameters.get_parameter_value(PName.MIN_FREQUENCY))
    max_freq    = cast(float, parameters.get_parameter_value(PName.MAX_FREQUENCY))
    freq_step   = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       freq_step)
    if num_steps_per_sweep <= 1:
        raise ValueError((f"We need strictly greater than one step per sweep. "
                          f"Computed {num_steps_per_sweep} step per sweep"))
    

def validate_sweep_interval(
        parameters: Parameters
) -> None: 
    """Ensure that the sweep interval is greater than the batch size.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the sweep interval is greater than the batch size.
    """
    min_freq         = cast(float, parameters.get_parameter_value(PName.MIN_FREQUENCY))
    max_freq         = cast(float, parameters.get_parameter_value(PName.MAX_FREQUENCY))
    freq_step        = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))
    batch_size       = cast(int, parameters.get_parameter_value(PName.BATCH_SIZE))
    sample_rate      = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    num_steps_per_sweep = _compute_num_steps_per_sweep(min_freq, 
                                                       max_freq, 
                                                       freq_step)
    num_samples_per_sweep = num_steps_per_sweep * samples_per_step
    sweep_interval = num_samples_per_sweep * 1/sample_rate
    if sweep_interval > batch_size:
        raise ValueError((f"Sweep interval must be less than the batch size. "
                          f"The computed sweep interval is {sweep_interval} [s], "
                          f"but the given batch size is {batch_size} [s]"))
    

def validate_num_samples_per_step(
        parameters: Parameters
) -> None:
    """Ensure that the number of samples per step is greater than the window size.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the window size is greater than the number of samples per step.
    """
    window_size      = cast(int, parameters.get_parameter_value(PName.WINDOW_SIZE))
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))

    if window_size >= samples_per_step:
        raise ValueError((f"Window size must be strictly less than the number of samples per step. "
                          f"Got window size {window_size} [samples], which is more than or equal "
                          f"to the number of samples per step {samples_per_step}"))
    

def validate_non_overlapping_steps(
        parameters: Parameters
) -> None:
    """Ensure that the stepped spectrograms are non-overlapping in the frequency domain.

    :param parameters: The parameters to be validated.
    :raises NotImplementedError: If the spectrograms overlap in the frequency domain.
    """
    
    freq_step   = cast(float, parameters.get_parameter_value(PName.FREQUENCY_STEP))
    sample_rate = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    if freq_step < sample_rate:
        raise NotImplementedError(f"SPECTRE does not yet support spectral steps overlapping in frequency. "
                                  f"Got frequency step {freq_step * 1e-6} [MHz] which is less than the sample "
                                  f"rate {sample_rate * 1e-6} [MHz]")
    

def validate_step_interval(
        parameters: Parameters,
        api_retuning_latency: float
) -> None:
    """Ensure that the time elapsed collecting samples at a fixed frequency is greater 
    than the empirically derived API retuning latency.

    :param parameters: The parameters to be validated.
    :param api_retuning_latency: The empirically derived API retuning latency (in seconds).
    :raises ValueError: If the time elapsed for a step is less than the API retuning latency.
    """
    samples_per_step = cast(int, parameters.get_parameter_value(PName.SAMPLES_PER_STEP))
    sample_rate      = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    step_interval = samples_per_step * 1/ sample_rate # [s]
    if step_interval < api_retuning_latency:
        raise ValueError(f"The computed step interval of {step_interval} [s] is of the order of empirically "
                         f"derived api latency {api_retuning_latency} [s]; you may experience undefined behaviour!")


def validate_sample_rate_with_master_clock_rate(
    parameters: Parameters,
) -> None:
    """Ensure that the master clock rate is an integer multiple of the sample rate.

    :param parameters: The parameters to be validated.
    :raises ValueError: If the master clock rate is not an integer multiple of the sample rate
    """
    master_clock_rate = cast(int, parameters.get_parameter_value(PName.MASTER_CLOCK_RATE))
    sample_rate       = cast(int, parameters.get_parameter_value(PName.SAMPLE_RATE))

    if master_clock_rate % sample_rate != 0:
        raise ValueError(f"The master clock rate of {master_clock_rate} [Hz] is not an integer "
                         f"multiple of the sample rate {sample_rate} [Hz].")


def validate_fixed_center_frequency(
    parameters: Parameters
) -> None:
    """Apply validators for capture config parameters describing fixed center frequency capture.

    :param parameters: The parameters to be validated.
    """
    validate_nyquist_criterion(parameters)
    validate_window(parameters)


def validate_swept_center_frequency(
    parameters: Parameters,
    api_retuning_latency: Optional[float] = None,
) -> None:
    """Apply validators for capture config parameters describing swept center frequency capture.

    :param parameters: The parameters to be validated.
    :param api_retuning_latency: The empirically derived API retuning latency. Defaults to None.
    """
    validate_nyquist_criterion(parameters)
    validate_window(parameters)
    validate_non_overlapping_steps(parameters)
    validate_num_steps_per_sweep(parameters)
    validate_num_samples_per_step(parameters)
    validate_sweep_interval(parameters)
    
    if api_retuning_latency is not None:
        validate_step_interval(parameters, api_retuning_latency)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, overload

from spectre_core.capture_configs import (
    CaptureTemplate, CaptureMode, Parameters, Bound, PName,
    get_base_capture_template, get_base_ptemplate, OneOf,
    validate_fixed_center_frequency, validate_swept_center_frequency
)
from .._base import BaseReceiver
from .._spec_names import SpecName


def get_pvalidator_fixed_center_frequency(
    sdrplay_receiver: BaseReceiver 
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_fixed_center_frequency(parameters)
    return pvalidator


def get_pvalidator_swept_center_frequency(
    sdrplay_receiver: BaseReceiver 
) -> Callable[[Parameters], None]:
    def pvalidator(parameters: Parameters) -> None:
        validate_swept_center_frequency(parameters,
                                        sdrplay_receiver.get_spec(SpecName.API_RETUNING_LATENCY))
    return pvalidator


def get_capture_template_fixed_center_frequency(
    sdrplay_receiver: BaseReceiver
) -> CaptureTemplate:
    
    capture_template = get_base_capture_template( CaptureMode.FIXED_CENTER_FREQUENCY )
    capture_template.add_ptemplate( get_base_ptemplate(PName.BANDWIDTH) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.IF_GAIN) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.RF_GAIN) )

    capture_template.set_defaults(
        (PName.BATCH_SIZE,            3.0),
        (PName.CENTER_FREQUENCY,      95800000),
        (PName.SAMPLE_RATE,           600000),
        (PName.BANDWIDTH,             600000),
        (PName.WINDOW_HOP,            512),
        (PName.WINDOW_SIZE,           1024),
        (PName.WINDOW_TYPE,           "blackman"),
        (PName.RF_GAIN,               -30),
        (PName.IF_GAIN,               -30)
    )   

    capture_template.add_pconstraint(
        PName.CENTER_FREQUENCY,
        [
            Bound(
                lower_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=sdrplay_receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=sdrplay_receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            OneOf(
                sdrplay_receiver.get_spec( SpecName.BANDWIDTH_OPTIONS )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.IF_GAIN,
        [
            Bound(
                upper_bound=sdrplay_receiver.get_spec(SpecName.IF_GAIN_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.RF_GAIN,
        [
            Bound(
                upper_bound=sdrplay_receiver.get_spec(SpecName.RF_GAIN_UPPER_BOUND)
            )
        ]
    )
    return capture_template


def get_capture_template_swept_center_frequency(
    sdrplay_receiver: BaseReceiver
) -> CaptureTemplate:

    capture_template = get_base_capture_template( CaptureMode.SWEPT_CENTER_FREQUENCY )
    capture_template.add_ptemplate( get_base_ptemplate(PName.BANDWIDTH) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.IF_GAIN) )
    capture_template.add_ptemplate( get_base_ptemplate(PName.RF_GAIN) )

    capture_template.set_defaults(
        (PName.BATCH_SIZE,            4.0),
        (PName.MIN_FREQUENCY,         95000000),
        (PName.MAX_FREQUENCY,         100000000),
        (PName.SAMPLES_PER_STEP,      80000),
        (PName.FREQUENCY_STEP,        1536000),
        (PName.SAMPLE_RATE,           1536000),
        (PName.BANDWIDTH,             1536000),
        (PName.WINDOW_HOP,            512),
        (PName.WINDOW_SIZE,           1024),
        (PName.WINDOW_TYPE,           "blackman"),
        (PName.RF_GAIN,               -30),
        (PName.IF_GAIN,               -30)
    )   

    capture_template.add_pconstraint(
        PName.MIN_FREQUENCY,
        [
            Bound(
                lower_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.MAX_FREQUENCY,
        [
            Bound(
                lower_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_LOWER_BOUND),
                upper_bound=sdrplay_receiver.get_spec(SpecName.FREQUENCY_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.SAMPLE_RATE,
        [
            Bound(
                lower_bound=sdrplay_receiver.get_spec(SpecName.SAMPLE_RATE_LOWER_BOUND),
                upper_bound=sdrplay_receiver.get_spec(SpecName.SAMPLE_RATE_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.BANDWIDTH,
        [
            OneOf(
                sdrplay_receiver.get_spec( SpecName.BANDWIDTH_OPTIONS )
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.IF_GAIN,
        [
            Bound(
                upper_bound=sdrplay_receiver.get_spec(SpecName.IF_GAIN_UPPER_BOUND)
            )
        ]
    )
    capture_template.add_pconstraint(
        PName.RF_GAIN,
        [
            Bound(
                upper_bound=sdrplay_receiver.get_spec(SpecName.RF_GAIN_UPPER_BOUND)
            )
        ]
    )
    return capture_template



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum

class SpecName(Enum):
    """A hardware specification name.
    
    :ivar FREQUENCY_LOWER_BOUND: The lower bound for the center frequency, in Hz.
    :ivar FREQUENCY_UPPER_BOUND: The upper bound for the center frequency, in Hz.
    :ivar SAMPLE_RATE_LOWER_BOUND: The lower bound for the sampling rate, in Hz.
    :ivar SAMPLE_RATE_UPPER_BOUND: The upper bound for the sampling rate, in Hz.
    :ivar BANDWIDTH_LOWER_BOUND: The lower bound for the bandwidth, in Hz.
    :ivar BANDWIDTH_UPPER_BOUND: The upper bound for the bandwidth, in Hz.
    :ivar BANDWIDTH_OPTIONS: The permitted bandwidths for the receiver, in Hz.
    :ivar IF_GAIN_UPPER_BOUND: The upper bound for the intermediate frequency gain, in dB.
    Negative values indicate attenuation.
    :ivar RF_GAIN_UPPER_BOUND: The upper bound for the radio frequency gain, in dB. 
    Negative values indicate attenuation.
    :ivar GAIN_UPPER_BOUND: The upper bound for the gain, in dB.
    :ivar WIRE_FORMATS: Supported data types transferred over the bus/network.
    :ivar MASTER_CLOCK_RATE_LOWER_BOUND:  The lower bound for the SDR reference clock rate, in Hz.
    :ivar MASTER_CLOCK_RATE_UPPER_BOUND:  The upper bound for the SDR reference clock rate, in Hz.
    :ivar API_RETUNING_LATENCY: An empirical estimate of the delay between issuing a command
    for a receiver to retune its center frequency and the actual physical update of the center frequency. 
    """
    FREQUENCY_LOWER_BOUND         = "frequency_lower_bound"
    FREQUENCY_UPPER_BOUND         = "frequency_upper_bound"
    SAMPLE_RATE_LOWER_BOUND       = "sample_rate_lower_bound"
    SAMPLE_RATE_UPPER_BOUND       = "sample_rate_upper_bound"
    BANDWIDTH_LOWER_BOUND         = "bandwidth_lower_bound"
    BANDWIDTH_UPPER_BOUND         = "bandwidth_upper_bound"
    BANDWIDTH_OPTIONS             = "bandwidth_options"
    IF_GAIN_UPPER_BOUND           = "if_gain_upper_bound"
    RF_GAIN_UPPER_BOUND           = "rf_gain_upper_bound"
    GAIN_UPPER_BOUND              = "gain_upper_bound"
    WIRE_FORMATS                  = "wire_formats"
    MASTER_CLOCK_RATE_LOWER_BOUND = "master_clock_rate_lower_bound"
    MASTER_CLOCK_RATE_UPPER_BOUND = "master_clock_rate_upper_bound"
    API_RETUNING_LATENCY          = "api_retuning_latency"



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Manage `spectre` jobs and workers."""

from ._jobs import Job, start_job
from ._workers import (
    Worker, make_worker, do_capture, do_post_processing
)

__all__ = [
    "Job", "Worker", "make_worker", "start_job", "do_capture", "do_post_processing"
]


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from typing import TypeVar
from functools import cached_property
from abc import ABC, abstractmethod

from spectre_core._file_io import BaseFileHandler
from spectre_core.config import get_batches_dir_path, TimeFormat
from spectre_core.spectrograms import Spectrogram


T = TypeVar('T')

class BatchFile(BaseFileHandler[T]):
    """Abstract base class for files belonging to a batch, identified by their file extension.

    Batch file names must conform to the following structure:

        `<start time>_<tag>.<extension>`

    The substring `<start time>_<tag>` is referred to as the batch name. Files with the same batch name 
    belong to the same batch.
    """
    def __init__(
        self, 
        batch_parent_dir_path: str, 
        batch_name: str, 
        extension: str
    ) -> None:
        """Initialise a `BatchFile` instance.

        :param batch_parent_dir_path: Parent directory of the batch.
        :param batch_name: Base file name, composed of the batch start time and tag.
        :param extension: File extension.
        """
        super().__init__(batch_parent_dir_path, 
                         batch_name, 
                         extension)
        self._start_time, self._tag = batch_name.split("_")
   
   
    @property
    def start_time(
        self
    ) -> str:
        """The start time of the batch, formatted as a string up to seconds precision."""
        return self._start_time


    @cached_property
    def start_datetime(
        self
    ) -> datetime:
        """The start time of the batch, parsed as a datetime up to seconds precision."""
        return datetime.strptime(self.start_time, TimeFormat.DATETIME)

    
    @property
    def tag(
        self
    ) -> str:
        """The batch name tag."""
        return self._tag
   
 
class BaseBatch(ABC):
    """
    An abstract base class representing a group of data files over a common time interval.

    All files in a batch share a base file name and differ only by their extension.
    Subclasses of `BaseBatch` define the expected data for each file extension and 
    provide an API for accessing their contents using `BatchFile` subclasses.

    Subclasses should expose `BatchFile` instances directly as attributes, which
    simplifies static typing. Additionally, they should call `add_file` in the constructor
    to formally register each `BatchFile`.
    """
    def __init__(
        self, 
        start_time: str,
        tag: str
    ) -> None:
        """Initialise a `BaseBatch` instance.

        :param start_time: Start time of the batch as a string with seconds precision.
        :param tag: The batch name tag.
        """
        self._start_time = start_time
        self._tag: str = tag
        self._start_datetime = datetime.strptime(self.start_time, TimeFormat.DATETIME)
        self._parent_dir_path = get_batches_dir_path(year  = self.start_datetime.year,
                                                     month = self.start_datetime.month,
                                                     day   = self.start_datetime.day)
        
        # internal register of batch files
        self._batch_files: dict[str, BatchFile] = {}
    
    
    @property
    @abstractmethod
    def spectrogram_file(
        self
    ) -> BatchFile:
        """The batch file which contains spectrogram data."""
        
    
    @property
    def start_time(
        self
    ) -> str:
        """The start time of the batch, formatted as a string up to seconds precision."""
        return self._start_time


    @property
    def start_datetime(
        self
    ) -> datetime:
        """The start time of the batch, parsed as a datetime up to seconds precision."""
        return self._start_datetime
    
    
    @property
    def tag(
        self
    ) -> str:
        """The batch name tag."""
        return self._tag
    

    @property
    def parent_dir_path(
        self
    ) -> str:
        """The parent directory for the batch."""
        return self._parent_dir_path


    @property
    def name(
        self
    ) -> str:
        """Return the base file name shared by all files in the batch, 
        composed of the start time and the batch tag."""
        return f"{self._start_time}_{self._tag}"
    
    
    @property
    def extensions(
        self
    ) -> list[str]:
        """All defined file extensions for the batch."""
        return list(self._batch_files.keys())
    
    
    @property
    def batch_files(
        self
    ) -> dict[str, BatchFile]:
        """Map each file extension in the batch to the corresponding batch file instance.
        
        Use `add_file` to add a file to the batch.
        """
        return self._batch_files
    

    def add_file(
        self, 
        batch_file: BatchFile
    ) -> None:
        """Add an instance of a batch file to the batch.

        :param batch_file: The `BatchFile` instance to add to the batch.
        :raises ValueError: If the `BatchFile` instance does not have a defined file extension.
        """
        if batch_file.extension is None:
            raise ValueError(f"The `BatchFile` must have a defined file extension. "
                             f"Received '{batch_file.extension}.")
        self._batch_files[batch_file.extension] = batch_file
    

    def get_file(
        self, 
        extension: str
    ) -> BatchFile:
        """Get a batch file instance from the batch, according to the file extension.

        :param extension: The file extension of the batch file.
        :raises NotImplementedError: If the extension is undefined for the batch.
        :return: The batch file instance registered under the input file extension.
        """
        try:
            return self._batch_files[extension]
        except KeyError:
            raise NotImplementedError(f"A batch file with extension '{extension}' is not implemented for this batch.")


    def delete_file(
        self, 
        extension: str
    ) -> None:
        """Delete a file from the batch, according to the file extension.

        :param extension: The file extension of the batch file.
        :raises FileNotFoundError: If the batch file does not exist in the file system.
        """
        batch_file = self.get_file(extension)
        batch_file.delete()


    def has_file(
        self, 
        extension: str
    ) -> bool:
        """Determine the existance of a batch file in the file system.

        :param extension: The file extension of the batch file.
        :return: True if the batch file exists in the file system, False otherwise.
        """
        try:
            batch_file = self.get_file(extension)
            return batch_file.exists
        except FileNotFoundError:
            return False
        
        
    def read_spectrogram(
        self
    ) -> Spectrogram:
        """Read and return the spectrogram data stored in the batch.

        :return: The spectrogram stored by the batch `spectrogram_file`.
        """
        return self.spectrogram_file.read()
    


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from abc import ABC, abstractmethod
from typing import TypeVar, Optional, Any, Generic

# value type
VT = TypeVar('VT')

class BasePConstraint(ABC, Generic[VT]):
    """An abstract base class for an arbitary parameter constraint."""
    @abstractmethod
    def constrain(
        self, 
        value: VT
    ) -> None:
        """Apply a constraint to the input parameter. Implementations must raise a `ValueError` for
        if the input value fails the constraint.

        :param value: The value to be constrained.
        """
        
        
    def __format__(
        self, 
        format_spec: str = ""
    ) -> str:
        attrs = ", ".join(f"{key}={value!r}" for key, value in vars(self).items())
        return f"{self.__class__.__name__}({attrs})"


class Bound(BasePConstraint[float | int]):
    """Bound a numeric parameter value to a some specified interval."""
    def __init__(
        self,
        lower_bound: Optional[float | int] = None,
        upper_bound: Optional[float | int] = None,
        strict_lower: bool = False,
        strict_upper: bool = False
    ) -> None:
        """Create an instance of `Bound`.

        :param lower_bound: The value must be greater than `lower_bound`. Inclusive if `strict_lower` is False. Defaults to None.
        :param upper_bound: The value must be less than `upper_bound`. Inclusive if `strict_upper` is False. Defaults to None.
        :param strict_lower: If true, the value must be strictly greater than `lower_bound`. Defaults to False.
        :param strict_upper: If true, the value must be strictly less than `upper_bound`. Defaults to False.
        """
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._strict_lower = strict_lower
        self._strict_upper = strict_upper

    def constrain(
        self, 
        value: float | int
    ) -> None:
        """Bound the parameter value to a specified interval.

        :param value: The value to be constrained.
        :raises ValueError: If the value is outside the specified interval.
        """
        if self._lower_bound is not None:
            if self._strict_lower and value <= self._lower_bound:
                raise ValueError(f"Value must be strictly greater than {self._lower_bound}. "
                                 f"Got {value}.")
            if not self._strict_lower and value < self._lower_bound:
                raise ValueError(f"Value must be greater than or equal to {self._lower_bound}. "
                                 f"Got {value}.")

        if self._upper_bound is not None:
            if self._strict_upper and value >= self._upper_bound:
                raise ValueError(f"Value must be strictly less than {self._upper_bound}. "
                                 f"Got {value}.")
            if not self._strict_upper and value > self._upper_bound:
                raise ValueError(f"Value must be less than or equal to {self._upper_bound}. "
                                 f"Got {value}.")

# option type
OT = TypeVar('OT')
class OneOf(BasePConstraint[OT]):
    """Constrain a parameter value to be one of a pre-defined list of options."""
    def __init__(
        self, 
        options: Optional[list[OT]] = None
    ) -> None:
        """Initialise an instance of `OneOf`.

        :param options: Input values are required to be one of `options`. If no options are provided,
        it is assumed to be an empty list. Defaults to None.
        """
        self._options = options or []


    def constrain(self, value: OT) -> None:
        """Constrain the input value to be one of a list of pre-defined options.

        :param value: The value to be constrained.
        :raises ValueError: If the input value is not one of the pre-defined options.
        """
        if value not in self._options:
            raise ValueError(f"Value must be one of {self._options}. Got {value}.")
        

class PowerOfTwo(BasePConstraint[int]):
    """Constrain a numeric parameter value to be a power of two."""
    def constrain(
        self, 
        value: int
    ) -> None:
        """Constrain the input value to be a power of two.

        :param value: The input value to be constrained.
        :raises ValueError: If the input value is not exactly some power of two.
        """
        if value <= 0 or (value & (value - 1)) != 0:
            raise ValueError(f"Value must be a power of two. Got {value}.")


@dataclass(frozen=True)
class EnforceSign:
    """Enforce the sign of some value.
    
    :ivar positive: Enforce the value to be strictly positive.
    :ivar negative: Enforce the value to be strictly negative.
    :ivar non_negative: Enforce the value to be zero or positive.
    :ivar non_positive: Enforce the value to be zero or negative.
    """
    positive       = Bound(lower_bound=0, strict_lower=True)
    negative       = Bound(upper_bound=0, strict_upper=True)
    non_negative   = Bound(lower_bound=0, strict_lower=False)
    non_positive   = Bound(upper_bound=0, strict_upper=False)



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional, TypeVar, Any, Generic, Callable
from textwrap import dedent
from copy import deepcopy

from ._pnames import PName
from ._pconstraints import BasePConstraint, EnforceSign, PowerOfTwo, Bound
from ._parameters import Parameter
    
# value type
VT = TypeVar('VT')
class PTemplate(Generic[VT]):
    """A template which constrains the value and type of a capture config parameter."""
    def __init__(
        self,
        name: PName,
        ptype: Callable[[Any], VT],
        default: Optional[VT] = None,
        nullable: bool = False,
        enforce_default: bool = False,
        help: Optional[str] = None,
        pconstraints: Optional[list[BasePConstraint]] = None
    ) -> None:
        """Initialise an instance of `PTemplate`.

        :param name: The name of the parameter.
        :param ptype: The required type of the parameter value.
        :param default: The parameter value if not explicitly specified. Defaults to None.
        :param nullable: Whether the value of the parameter can be `None`. Defaults to False.
        :param enforce_default: Indicates whether the value must match the specified `default`. Defaults to False.
        :param help: A helpful description of what the parameter is and the value it stores. Defaults to None.
        :param pconstraints: Custom constraints to be applied to the value of the parameter. Defaults to None.
        """
        self._name = name
        self._ptype = ptype
        self._default = default
        self._nullable = nullable
        self._enforce_default = enforce_default
        self._help = dedent(help).strip().replace("\n", " ") if help else "No help has been provided."
        self._pconstraints: list[BasePConstraint] = pconstraints or []


    @property
    def name(
        self
    ) -> PName:
        """The name of the parameter."""
        return self._name
    

    @property
    def ptype(
        self
    ) -> Callable[[object], VT]:
        """The required type of the parameter. The value must be castable as this type."""
        return self._ptype
    

    @property
    def default(
        self
    ) -> Optional[VT]:
        """The parameter value if not explictly specified."""
        return self._default
    

    @default.setter
    def default(
        self, 
        value: VT
    ) -> None:
        """Update the `default` of this parameter template.

        :param value: The new default value to set.
        """
        self._default = self._cast(value)


    @property
    def nullable(
        self
    ) -> bool:
        """Whether the value of the parameter can be `None`."""
        return self._nullable
    

    @property
    def enforce_default(
        self
    ) -> bool:
        """Indicates whether the value must match the specified `default`."""
        return self._enforce_default
    

    @enforce_default.setter
    def enforce_default(
        self, 
        value: bool
    ) -> None:
        """Update whether to `enforce_default` for this parameter template.

        :param value: Whether to enforce the default value.
        """
        self._enforce_default = value
    

    @property
    def help(
        self
    ) -> str:
        """A helpful description of what the parameter is, and the value it stores."""
        return self._help
    
    
    def add_pconstraint(
        self,
        pconstraint: BasePConstraint
    ) -> None:
        """Add a parameter constraint to this template.

        :param pconstraint: A `PConstraint` instance compatible with the `ptype`.
        """
        self._pconstraints.append(pconstraint)


    def _cast(
        self, 
        value: Any
    ) -> VT:
        """Cast the input value to the `ptype` for this parameter template.

        :param value: The value to be type casted.
        :raises ValueError: If there is any trouble casting `value` as the `ptype` for this parameter template.
        :return: The input value cast as `ptype` for this parameter template.
        """
        try:
            return self._ptype(value)
        except (TypeError, ValueError) as e:
            raise ValueError(f"Could not cast '{value}' to '{self._ptype.__name__}': {e}")


    def _constrain(
        self, 
        value: VT
    ) -> VT:
        """Constrain the input value according to constraints of the template.

        :param value: The value to be constrained.
        :raises ValueError: If a custom `PConstraint` fails for the input value.
        :raises RuntimeError: If an unexpected error occurs during constraint validation.
        :return: The input value unchanged if it passes validation.
        """
        if self._enforce_default and value != self._default:
            raise ValueError(f"The default value of '{self._default}' "
                             f"is required for the parameter '{self._name}'.")

        # apply existing pconstraints
        for constraint in self._pconstraints:
            try:
                constraint.constrain(value)
            except ValueError as e:
                raise ValueError(f"PConstraint '{constraint.__class__.__name__}' failed for the parameter '{self._name}': {e}")
            except Exception as e:
                raise RuntimeError(f"An unexpected error occurred while applying the pconstraint '{constraint.__class__.__name__}' to "
                                    f"'{self.name}': {e}")
        return value


    def apply_template(
        self, 
        value: Optional[Any]
    ) -> Optional[VT]:
        """Cast a value and validate it according to this parameter template.

        :param value: The input value.
        :raises ValueError: If the value is `None`, no `default` is specified, and the parameter is not nullable.
        :return: The input value type cast and validated according to the parameter template.
        """
        if value is None:
            if self._default is not None:
                value = self._default
            elif not self._nullable:
                raise ValueError(f"The parameter '{self._name}' is not nullable, "
                                 f"but no value or default has been provided. "
                                 f"Either provide a value, or provide a default.")
            else:
                return None
        
        return self._constrain( self._cast(value) )


    def make_parameter(
        self, 
        value: Optional[Any] = None
    ) -> Parameter:
        """Create a `Parameter` compliant with this template.

        :param value: The provided value for the parameter. Defaults to None.
        :return: A `Parameter` object validated according to this template.
        """
        value = self.apply_template(value)
        return Parameter(self._name, value)


    def to_dict(
        self
    ) -> dict[str, str]:
        """Convert this parameter template to a serialisable dictionary.

        :return: A dictionary representation of this parameter template with string formatted values.
        """
        d = {
            "name": self._name.value,
            "type": self._ptype.__name__,
            "default": self._default,
            "enforce_default": self._enforce_default,
            "help": self._help,
            "pconstraints": [f"{constraint}" for constraint in self._pconstraints]
        }
        return {k: f"{v}" for k,v in d.items()}
 

# ------------------------------------------------------------------------------------------ #
# `_base_ptemplates` holds all shared base parameter templates. They are 'base' templates, 
# in the sense that they should be configured according to specific use-cases. For example, 
# `default` values should be set, and `pconstraints` added according to specific SDR specs.
# ------------------------------------------------------------------------------------------ # 

_base_ptemplates: dict[PName, PTemplate] = {
    PName.CENTER_FREQUENCY:       PTemplate(PName.CENTER_FREQUENCY,       
                                            float, 
                                            help = """
                                                The center frequency of the SDR in Hz.
                                                This value determines the midpoint of the frequency range
                                                being processed.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.MIN_FREQUENCY:          PTemplate(PName.MIN_FREQUENCY,          
                                            float, 
                                            help = """
                                                The minimum center frequency, in Hz, for the frequency sweep.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.MAX_FREQUENCY:          PTemplate(PName.MAX_FREQUENCY,          
                                            float, 
                                            help = """
                                                The maximum center frequency, in Hz, for the frequency sweep.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.FREQUENCY_STEP:         PTemplate(PName.FREQUENCY_STEP,         
                                            float, 
                                            help = """
                                                The amount, in Hz, by which the center frequency is incremented
                                                for each step in the frequency sweep. 
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.BANDWIDTH:              PTemplate(PName.BANDWIDTH,              
                                            float, 
                                            help = """
                                                The frequency range in Hz the signal will occupy without
                                                significant attenutation.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.SAMPLE_RATE:            PTemplate(PName.SAMPLE_RATE,            
                                            int,   
                                            help = """
                                                The number of samples per second in Hz.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.IF_GAIN:                PTemplate(PName.IF_GAIN,                
                                            float, 
                                            help = """
                                                The intermediate frequency gain, in dB.
                                                Negative value indicates attenuation.
                                                """,
                                            pconstraints=[
                                                EnforceSign.negative
                                                ]),
    PName.RF_GAIN:                PTemplate(PName.RF_GAIN,                
                                            float, 
                                            help = """
                                                The radio frequency gain, in dB.
                                                Negative value indicates attenuation.
                                                """,
                                            pconstraints=[
                                                EnforceSign.non_positive
                                                ]),
    PName.GAIN:                   PTemplate(PName.GAIN,
                                            float,
                                            help = """
                                                 The gain value for the SDR, in dB 
                                                 """
                                           ),
    PName.MASTER_CLOCK_RATE:      PTemplate(PName.MASTER_CLOCK_RATE,
                                            int,
                                            help = """
                                                 The primary reference clock for the SDR, specified in Hz. 
                                                 """
                                           ),
    PName.WIRE_FORMAT:            PTemplate(PName.WIRE_FORMAT,
                                            str,
                                            help = """
                                                 Controls the form of the data over the bus/network.
                                                 """
                                            ),
    PName.EVENT_HANDLER_KEY:      PTemplate(PName.EVENT_HANDLER_KEY,      
                                            str,
                                            help = """
                                                Identifies which post-processing functions to invoke
                                                on newly created files.
                                                """),
    PName.BATCH_KEY:              PTemplate(PName.BATCH_KEY,              
                                            str,
                                            help = """
                                                Identifies the type of data is stored in each batch.
                                                """,
                                            ),
    PName.WINDOW_SIZE:            PTemplate(PName.WINDOW_SIZE,            
                                            int,  
                                            help = """
                                                The size of the window, in samples, when performing the
                                                Short Time FFT.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive, 
                                                PowerOfTwo(),
                                                ]),
    PName.WINDOW_HOP:             PTemplate(PName.WINDOW_HOP,             
                                            int,   
                                            help = """
                                                How much the window is shifted, in samples, 
                                                when performing the Short Time FFT.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.WINDOW_TYPE:            PTemplate(PName.WINDOW_TYPE,            
                                            str,
                                            help = """
                                                The type of window applied when performing the Short
                                                Time FFT.
                                                """,
                                            ),
    PName.WATCH_EXTENSION:        PTemplate(PName.WATCH_EXTENSION,        
                                            str,
                                            help = """
                                                Post-processing is triggered by newly created files with this extension.
                                                Extensions are specified without the '.' character.
                                                """,
                                            ),
    PName.TIME_RESOLUTION:        PTemplate(PName.TIME_RESOLUTION,        
                                            float, 
                                            nullable=True,
                                            help = """
                                                Batched spectrograms are smoothed by averaging up to the time resolution,
                                                specified in seconds.
                                                """,
                                            pconstraints=[
                                                EnforceSign.non_negative
                                                ]),
    PName.FREQUENCY_RESOLUTION:   PTemplate(PName.FREQUENCY_RESOLUTION,   
                                            float, 
                                            nullable=True,
                                            help = """
                                                Batched spectrograms are smoothed by averaging up to the frequency resolution,
                                                specified in Hz.
                                                """,
                                            pconstraints=[
                                                EnforceSign.non_negative
                                                ]),
    PName.TIME_RANGE:             PTemplate(PName.TIME_RANGE,             
                                            float, 
                                            nullable=True,
                                            help = """
                                                Batched spectrograms are stitched together until
                                                the time range, in seconds, is surpassed.
                                                """,
                                            pconstraints=[
                                                EnforceSign.non_negative
                                                ]),
    PName.BATCH_SIZE:             PTemplate(PName.BATCH_SIZE,             
                                            int,   
                                            help = """
                                                SDR data is collected in batches of this size, specified
                                                in seconds.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.SAMPLES_PER_STEP:       PTemplate(PName.SAMPLES_PER_STEP,       
                                            int,   
                                            help = """
                                                The number of samples taken at each center frequency in the sweep.
                                                This may vary slightly from what is specified due to the nature of
                                                GNU Radio runtime.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.ORIGIN:                 PTemplate(PName.ORIGIN,
                                            str,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword ORIGIN.
                                                """),
    PName.TELESCOPE:              PTemplate(PName.TELESCOPE,
                                            str,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword TELESCOP.
                                                """),
    PName.INSTRUMENT:             PTemplate(PName.INSTRUMENT,
                                            str,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword INSTRUME.
                                                """),
    PName.OBJECT:                 PTemplate(PName.OBJECT,
                                            str,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword OBJECT.
                                                """),
    PName.OBS_LAT:                PTemplate(PName.OBS_LAT,
                                            float,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword OBS_LAT.
                                                """),
    PName.OBS_LON:                PTemplate(PName.OBS_LON,
                                            float,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword OBS_LONG.
                                                """),
    PName.OBS_ALT:                PTemplate(PName.OBS_ALT,
                                            float,
                                            nullable=True,
                                            help="""
                                                Corresponds to the FITS keyword OBS_ALT.
                                                """),
    PName.AMPLITUDE:              PTemplate(PName.AMPLITUDE,
                                            float,
                                            help="""
                                                The amplitude of the signal.
                                                """),
    PName.FREQUENCY:              PTemplate(PName.FREQUENCY,
                                            float,
                                            help="""
                                                The frequency of the signal, in Hz.
                                                """),
    PName.MIN_SAMPLES_PER_STEP:   PTemplate(PName.MIN_SAMPLES_PER_STEP,
                                            int,
                                            help="""
                                                The number of samples in the shortest step of the staircase.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.MAX_SAMPLES_PER_STEP:   PTemplate(PName.MAX_SAMPLES_PER_STEP,
                                            int,
                                            help="""
                                                The number of samples in the longest step of the staircase.
                                                """,
                                        pconstraints=[
                                                EnforceSign.positive
                                                ]),
    PName.STEP_INCREMENT:         PTemplate(PName.STEP_INCREMENT,
                                            int,
                                            help="""
                                                The length by which each step in the staircase is incremented.
                                                """,
                                            pconstraints=[
                                                EnforceSign.positive,
                                            ])
                                    
                                           
}


def get_base_ptemplate(
    pname: PName,
) -> PTemplate:
    """Get a pre-defined base parameter template, to be configured according to the specific use case.

    :param pname: The parameter name for the template.
    :raises KeyError: If there is no base parameter template corresponding to the input name.
    :return: A deep copy of the corresponding base parameter template, if it exists.
    """
    if pname not in _base_ptemplates:
        raise KeyError(f"No ptemplate found for the parameter name '{pname}'. "
                       f"Expected one of {list(_base_ptemplates.keys())}")
    # A deep copy is required as each receiver instance may mutate the original instance
    # according to its particular use case. Copying preserves the original instance,
    # enabling reuse.
    return deepcopy( _base_ptemplates[pname] )



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum

class ProcessType(Enum):
    """The origin of a `spectre` process.
    
    :ivar USER: A process is one initiated directly by the user, or part of the main user session.
    :ivar WORKER: A process is one which is created and managed internally by `spectre`.
    """
    USER   = "user"
    WORKER = "worker"


# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum

class PName(Enum):
    """A `spectre` capture config parameter name.
    
    Each `PName` has an associated base parameter template which can be fetched
    using:
    
    `get_base_ptemplate`
    
    All parameters and parameter templates must take on some name in `PName`.
    To introduce a new parameter/parameter template, you need to create a new `PName`
    constant.
    """
    CENTER_FREQUENCY     = "center_frequency"
    MIN_FREQUENCY        = "min_frequency"
    MAX_FREQUENCY        = "max_frequency"
    FREQUENCY_STEP       = "frequency_step"
    FREQUENCY            = "frequency"
    BANDWIDTH            = "bandwidth"
    SAMPLE_RATE          = "sample_rate"
    IF_GAIN              = "if_gain"
    RF_GAIN              = "rf_gain"
    AMPLITUDE            = "amplitude"
    TIME_RESOLUTION      = "time_resolution"
    FREQUENCY_RESOLUTION = "frequency_resolution"
    TIME_RANGE           = "time_range"
    BATCH_SIZE           = "batch_size"
    WINDOW_TYPE          = "window_type"
    WINDOW_HOP           = "window_hop"
    WINDOW_SIZE          = "window_size"
    EVENT_HANDLER_KEY    = "event_handler_key"
    WATCH_EXTENSION      = "watch_extension"
    BATCH_KEY            = "batch_key"
    SAMPLES_PER_STEP     = "samples_per_step"
    MIN_SAMPLES_PER_STEP = "min_samples_per_step"
    MAX_SAMPLES_PER_STEP = "max_samples_per_step"
    STEP_INCREMENT       = "step_increment"
    ORIGIN               = "origin"
    TELESCOPE            = "telescope"
    INSTRUMENT           = "instrument"
    OBJECT               = "object"
    OBS_LAT              = "obs_lat"
    OBS_LON              = "obs_lon"
    OBS_ALT              = "obs_alt"
    GAIN                 = "gain"
    MASTER_CLOCK_RATE    = "master_clock_rate"
    WIRE_FORMAT          = "wire_format"



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Type, Callable, TypeVar

from .plugins._event_handler_keys import EventHandlerKey
from ._base import BaseEventHandler

# Map populated at runtime via the `register_event_handler` decorator.
event_handler_map: dict[EventHandlerKey, Type[BaseEventHandler]] = {}

T = TypeVar('T', bound=BaseEventHandler)
def register_event_handler(
    event_handler_key: EventHandlerKey
) -> Callable[[Type[T]], Type[T]]:
    """Decorator to register a `BaseEventHandler` subclass under a specified `EventHandlerKey`.

    :param event_handler_key: The key to register the `BaseEventHandler` subclass.
    :raises ValueError: If the provided `event_handler_key` is already registered.
    :return: A decorator that registers the `BaseBatch` subclass under the given `batch_key`.
    """
    def decorator(
        cls: Type[T]
    ) -> Type[T]:
        if event_handler_key in event_handler_map:
            raise ValueError(f"An event handler with key '{event_handler_key}' is already registered!")
        event_handler_map[event_handler_key] = cls
        return cls
    return decorator




# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from abc import ABC, abstractmethod
from typing import Callable, Optional, Literal, overload, Any

from spectre_core.exceptions import ModeNotFoundError
from spectre_core.capture_configs import (
    CaptureTemplate, Parameters, CaptureConfig
)
from .plugins._receiver_names import ReceiverName
from ._spec_names import SpecName

class BaseReceiver(ABC):
    """Abstract base class for software-defined radio receivers."""
    def __init__(
        self, 
        name: ReceiverName, 
        mode: Optional[str] = None
    ) -> None:
        """Initialise an instance of `BaseReceiver`.

        :param name: The name of the receiver.
        :param mode: The initial active operating mode. Defaults to None.
        """
        self._name = name

        self._specs: dict[SpecName, float|int|list[float|int]] = {}
        self._add_specs()
        
        self._capture_methods: dict[str, Callable[[str, Parameters], None]] = {}
        self._add_capture_methods()
        
        self._capture_templates: dict[str, CaptureTemplate] = {}
        self._add_capture_templates()
    
        self._pvalidators: dict[str, Callable[[Parameters], None]] = {}
        self._add_pvalidators()

        self._mode = None
        if mode is not None:
            self.mode = mode


    @abstractmethod
    def _add_specs(
        self
    ) -> None:
        """Subclasses must use `add_spec` to add hardware specifications."""
    

    @abstractmethod
    def _add_capture_methods(
        self
    ) -> None:
        """Subclasses must use `add_capture_method` to specify which method is called to capture 
        data, for each operating mode."""
        

    @abstractmethod
    def _add_capture_templates(
        self
    ) -> None:
        """Subclasses must use `add_capture_template` to define a `CaptureTemplate` for each operating mode."""


    @abstractmethod
    def _add_pvalidators(
        self
    ) -> None:
        """Subclasses must use `add_pvalidator` to add a parameter validation function (pvalidator) 
        for each operating mode."""

    
    @property
    def name(
        self
    ) -> ReceiverName:
        """The name of the receiver."""
        return self._name
    

    @property
    def capture_methods(
        self
    ) -> dict[str, Callable[[str, Parameters], None]]:
        """For each operating mode, the method which is called to capture data."""
        return self._capture_methods
  
  
    @property
    def capture_templates(
        self
    ) -> dict[str, CaptureTemplate]:
        """For each operating mode, the corresponding `CaptureTemplate`."""
        return self._capture_templates  


    @property
    def pvalidators(
        self
    ) -> dict[str, Callable[[Parameters], None]]:
        """For each operating mode, the corresponding parameter validation function (pvalidator)."""
        return self._pvalidators


    @property
    def specs(
        self
    ) -> dict[SpecName, float|int|list[float|int]]:
        """The hardware specifications."""
        return self._specs


    @property
    def modes(
        self
    ) -> list[str]:
        """The operating modes for the receiver.

        :raises ValueError: If the modes are inconsistent between `capture_methods`,
        `pvalidators` and `capture_templates`.
        """
        capture_method_modes    = list(self.capture_methods.keys())
        pvalidator_modes        = list(self.pvalidators.keys())
        capture_template_modes  = list(self.capture_templates.keys())
        
        if not capture_method_modes == pvalidator_modes == capture_template_modes:
            raise ValueError(f"Mode mismatch for the receiver {self.name}.")
        return capture_method_modes


    @property
    def mode(
        self
    ) -> str:
        """The active operating mode for the receiver."""
        if self._mode is None:
            raise ValueError(f"The operating mode for the receiver `{self.name.value}` is not set.")
        return self._mode


    @mode.setter
    def mode(
        self, 
        value: str,
    ) -> None:
        """Set the active operating mode.

        :param value: The new operating mode to activate.
        :raises ModeNotFoundError: If the specified mode is not defined in `modes`.
        """
        if (value not in self.modes):
            raise ModeNotFoundError((f"{value} is not a defined mode for the receiver {self.name}. "
                                     f"Expected one of {self.modes}"))
        self._mode = value


    @property
    def capture_method(
        self
    ) -> Callable[[str, Parameters], None]:
        """Start capturing data under the active operating mode."""
        return self.capture_methods[self.mode]


    @property
    def pvalidator(
        self
    ) -> Callable[[Parameters], None]:
        """The parameter validation function for the active operating mode."""
        return self.pvalidators[self.mode]


    @property
    def capture_template(
        self
    ) -> CaptureTemplate:
        """The `CaptureTemplate` for the active operating mode."""
        return self._capture_templates[self.mode]


    def add_capture_method(
        self,
        mode: str,
        capture_method: Callable[[str, Parameters], None]
    ) -> None:
        """
        Add a capture method for a specific operating mode.

        :param mode: The operating mode.
        :param capture_method: The function which captures data.
        """
        self._capture_methods[mode] = capture_method


    def add_pvalidator(
        self,
        mode: str,
        pvalidator: Callable[[Parameters], None]
    ) -> None:
        """
        Add a parameter validation function for a specific operating mode.

        :param mode: The operating mode.
        :param pvalidator: The validation function.
        """
        self._pvalidators[mode] = pvalidator


    def add_capture_template(
        self,
        mode: str,
        capture_template: CaptureTemplate
    ) -> None:
        """
        Add a `CaptureTemplate` for a specific operating mode.

        :param mode: The operating mode.
        :param capture_template: The capture template, defining parameter requirements.
        """
        self._capture_templates[mode] = capture_template


    def add_spec(
        self,
        name: SpecName,
        value: Any 
    ) -> None:
        """
        Add a hardware specification.

        :param name: The specification's name.
        :param value: The specification's value.
        """
        self.specs[name] = value

    
    def get_spec(
        self, 
        spec_name: SpecName
    ) -> Any:
        """
        Retrieve a hardware specification.

        :param spec_name: The name of the specification.
        :raises KeyError: If the specification is not found.
        :return: The specification's value.
        """
        if spec_name not in self.specs:
            raise KeyError(f"Spec not found with name '{spec_name}' "
                           f"for the receiver '{self.name}'")
        return self.specs[spec_name]


    def start_capture(
        self, 
        tag: str
    ) -> None:
        """Start capturing data in the active operating mode.

        :param tag: The tag of the capture config to load.
        """
        self.capture_method( tag, self.load_parameters(tag) )


    def save_parameters(
        self,
        tag: str,
        parameters: Parameters,
        force: bool = False
    ) -> None:
        """Create a capture config according to the active operating mode and save the 
        input parameters.

        The input parameters are validated before being written to file.
        
        :param tag: The tag identifying the capture config.
        :param parameters: The parameters to save in the capture config.
        :param force: If True, overwrites the existing file if it already exists. Defaults to False.
        """
        parameters = self.capture_template.apply_template(parameters)
        self.pvalidator(parameters)

        capture_config = CaptureConfig(tag)
        capture_config.save_parameters(self.name.value,
                                       self.mode,
                                       parameters,
                                       force)

    def load_parameters(
        self,
        tag: str
    ) -> Parameters:
        """Load a capture config, and return the parameters it stores.

        The parameters are validated before being returned.

        :param tag: The tag identifying the capture config.
        :return: The validated parameters stored in the capture config.
        """
        capture_config = CaptureConfig(tag)

        parameters = self.capture_template.apply_template(capture_config.parameters)
        self.pvalidator(parameters)
        
        return parameters



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from logging import getLogger
_LOGGER = getLogger(__name__)

import numpy as np
from typing import Tuple, cast
from datetime import timedelta

import os
import numpy.typing as npt
import numpy as np

from spectre_core.capture_configs import CaptureConfig, PName
from spectre_core.batches import IQStreamBatch
from spectre_core.spectrograms import Spectrogram, SpectrumUnit, time_average, frequency_average
from ._event_handler_keys import EventHandlerKey
from .._base import BaseEventHandler, make_sft_instance
from .._register import register_event_handler


def _do_stfft(
    iq_data: npt.NDArray[np.complex64],
    capture_config: CaptureConfig,
) -> Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """Do a Short-time Fast Fourier Transform on an array of complex IQ samples.
    
    The computation requires extra metadata, which is extracted from the detached header in the batch 
    and the capture config used to capture the data.
    
    The current implementation relies heavily on the `ShortTimeFFT` implementation from 
    `scipy.signal` (https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.ShortTimeFFT.html)
    which takes up a lot of the compute time.
    """

    sft = make_sft_instance(capture_config)

    # set p0=0, since by convention in the STFFT docs, p=0 corresponds to the slice centred at t=0
    p0=0

    # set p1 to the index of the first slice where the "midpoint" of the window is still inside the signal
    num_samples = len(iq_data)
    p1 = sft.upper_border_begin(num_samples)[1]
    
    # compute a ShortTimeFFT on the IQ samples
    complex_spectra = sft.stft(iq_data, 
                               p0 = p0, 
                               p1 = p1) 
    
    # compute the magnitude of each spectral component
    dynamic_spectra = np.abs(complex_spectra)


    # assign a physical time to each spectrum
    # p0 is defined to correspond with the first sample, at t=0 [s]
    times = sft.t(num_samples, 
                  p0 = p0, 
                  p1 = p1)
    # assign physical frequencies to each spectral component
    frequencies = sft.f + cast(float, capture_config.get_parameter_value(PName.CENTER_FREQUENCY))

    return times.astype(np.float32), frequencies.astype(np.float32), dynamic_spectra.astype(np.float32)


def _build_spectrogram(
    batch: IQStreamBatch,
    capture_config: CaptureConfig
) -> Spectrogram:
    """Generate a spectrogram using `IQStreamBatch` IQ samples."""
    # read the data from the batch
    iq_metadata = batch.hdr_file.read()
    iq_samples  = batch.bin_file.read()

    times, frequencies, dynamic_spectra = _do_stfft(iq_samples,
                                                    capture_config)

    # compute the start datetime for the spectrogram by adding the millisecond component to the batch start time
    spectrogram_start_datetime = batch.start_datetime + timedelta(milliseconds=iq_metadata.millisecond_correction)
    
    return Spectrogram(dynamic_spectra,
                       times,
                       frequencies,
                       batch.tag,
                       SpectrumUnit.AMPLITUDE,
                       spectrogram_start_datetime)


@register_event_handler(EventHandlerKey.FIXED_CENTER_FREQUENCY)
class FixedEventHandler(BaseEventHandler):
    def process(
        self, 
        absolute_file_path: str
    ) -> None:
        """Compute a spectrogram using `IQStreamBatch` IQ samples, cache it, then save it to file in the FITS
        format. The IQ samples are assumed to have been collected at a fixed center frequency.
        
        The computed spectrogram is averaged in time and frequency as per the user-configured capture config.
        Once the spectrogram has been computed successfully, the `.bin` and `.hdr` files are removed.

        :param absolute_file_path: The absolute file path of the `.bin` file in the batch.
        """
        _LOGGER.info(f"Processing: {absolute_file_path}")
        file_name = os.path.basename(absolute_file_path)
        base_file_name, _ = os.path.splitext(file_name)
        batch_start_time, tag = base_file_name.split('_')

        batch = IQStreamBatch(batch_start_time, tag)

        _LOGGER.info("Creating spectrogram")
        spectrogram = _build_spectrogram(batch,
                                         self._capture_config)
        
        time_resolution = cast(float, self._capture_config.get_parameter_value(PName.TIME_RESOLUTION))
        spectrogram = time_average(spectrogram,
                                   resolution=time_resolution)

        frequency_resolution = cast(float, self._capture_config.get_parameter_value(PName.FREQUENCY_RESOLUTION))
        spectrogram = frequency_average(spectrogram,
                                        resolution=frequency_resolution)
        
        self._cache_spectrogram(spectrogram)

        _LOGGER.info(f"Deleting {batch.bin_file.file_path}")
        batch.bin_file.delete()

        _LOGGER.info(f"Deleting {batch.hdr_file.file_path}")
        batch.hdr_file.delete()



# SPDX-FileCopyrightText: © 2024 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum

class ReceiverName(Enum):  
    """A `spectre` supported receiver.
    
    :ivar RSP1A: SDRPlay RSP1A
    :ivar RSPDUO: SDRPlay RSPduo
    :ivar TEST: `spectre` test receiver.
    :ivar B200MINI: USRP B200mini.
    """
    RSP1A     = "rsp1a"
    RSPDUO    = "rspduo"
    TEST      = "test"
    B200MINI =  "b200mini"
